#!/usr/bin/env python3
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level :mod:`PySide2`-based application-wide signal classes.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QObject, Signal

# ....................{ CLASSES                            }....................
class QBetseeSignaler(QObject):
    '''
    :class:`PySide2`-based collection of various application-wide signals.

    These signals permit callers to trigger handling of events by corresponding
    slots of interested objects and widgets distributed throughout the
    application, including:

    * Restoration and storage of application-wide settings to and from their
      on-disk backing store (e.g., an application- and user-specific dotfile).

    Design
    ----------
    This class has been intentionally isolated from all sibling classes (e.g.,
    :class:`QBetseeSettings`) to circumvent circular chicken-and-the-egg issues
    between this and the :class:`QBetseeMainWindow` class. Conjoining these
    sibling classes into one monolithic class would introduce non-trivial (and
    probably non-resolvable) complications, including the need for the conjoined
    class to retain a weak reference to its :class:`QBetseeMainWindow` parent,
    which could conceivably be prematurely destroyed by Qt in another thread.
    '''

    # ..................{ SIGNALS ~ settings                 }..................
    restore_settings_signal = Signal()
    '''
    Signal reading and restoring application-wide settings previously written to
    a predefined application- and user-specific on-disk file by the most recent
    execution of this application if any *or* reducing to a noop.

    This signal is connected to the
    :meth:`QBetseeSettings.restore_settings` slot at initialization time,
    enabling callers in different threads to thread-safely restore settings.
    '''


    store_settings_signal = Signal()
    '''
    Signal writing application-wide settings to a predefined application- and
    user-specific on-disk file, which the next execution of this application
    will read and restore on startup.

    This signal is connected to the
    :meth:`QBetseeSettings.store_settings` slot at initialization time, enabling
    callers in different threads to thread-safely store settings.
    '''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **pooled worker** (i.e., thread-safe object implementing generically
startable, pausable, resumable, and stoppable business logic isolated to a
dedicated thread by a parent :class:`QThreadPool` container) classes.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import (
    # QCoreApplication,
    QMutex,
    QMutexLocker,
    QRunnable,
    QWaitCondition,
)
from betse.exceptions import BetseMethodUnimplementedException
# from betse.util.io.log import logs
from betse.util.py import pyref
from betse.util.type.types import (
    type_check,
    CallableTypes,
    CallableOrNoneTypes,
    MappingOrNoneTypes,
    SequenceOrNoneTypes,
)
from betsee.guiexception import (
    BetseePySideThreadWorkerException,
    BetseePySideThreadWorkerStopException,
)
from betsee.util.thread import guithread
from betsee.util.thread.guithreadenum import ThreadWorkerState
from betsee.util.thread.pool.guipoolworksig import (
    QBetseeThreadPoolWorkerSignals,
)
from betsee.util.type.guitype import (
    QBetseeProgressBarOrNoneTypes,
    QLabelOrNoneTypes,
)

# ....................{ GLOBALS                           }....................
_worker_id_next = 0
'''
Non-thread-safe 0-based integer uniquely identifying the next **pooled worker**
(i.e., instance of the :class:`QBetseeThreadPoolWorker` superclass).
'''


_worker_id_lock = QMutex()
'''
Non-exception-safe mutual exclusion primitive rendering the
:func:`_get_worker_id_next` function thread-safe. This primitive is
non-exception-safe and hence should *never* be accessed directly. Each access
to this primitive should be encapsulated by instantiating an exception-safe
one-off :class:QMutexLocker` context manager as the target of a `with` context.
Note that the context provided by the :class:QMutexLocker` class is *not*
safely reusable and hence *must* be re-instantiated in each ``with`` context.
'''

# ....................{ SUPERCLASSES                      }....................
class QBetseeThreadPoolWorker(QRunnable):
    '''
    Low-level **pooled worker** (i.e., thread-safe object implementing
    generically startable, pausable, resumable, and stoppable business logic
    isolated to a dedicated thread by a parent :class:`QThreadPool` container).

    Lifecycle
    ----------
    The caller remains responsible for the lifecycle of this worker, an
    intentional design decision with several notable implications:

    * The caller *must* preserve a reference to this worker for the duration of
      the work performed by this worker (i.e., the :meth:`run` method), ideally
      as an instance variable of the class calling the
      :func:`guipoolthread.start_worker` function directing this work.
      Conversely, a local variable fails to satisfy this requirement. Lastly,
      note that preserving a reference to this reference is effectively
      mandatory for other unrelated reasons: notably, pausing, resuming, and
      stopping work by calling those methods on the reference to this worker.
    * If this is a **one-shot worker** (i.e., run only once by a single thread)
      rather than **recyclable worker** (i.e., run multiple times by multiple
      threads), the caller should connect the :attr:`signals.finished` signal
      of this worker to a slot nullifying the aforementioned reference to this
      worker. Failure to do so will result in a minor memory leak.

    By Qt default, workers are implicitly deleted by their parent
    :class:`QThreadPool` container immediately on returning from the
    :meth:`run` method. Although a sensible default for C++-based workers for
    whom lifecycle management is a non-trivial concern, Python-based workers
    suffer no such constraints and introduces numerous issues; in particular,
    allowing Qt to implicitly delete workers on worker completion:

    * Prevents callers from optionally recycling otherwise recyclable workers.
    * Introduces race conditions between the continued usage of these workers
      from Python callables from differing threads (e.g., slots, event
      handlers) and the deletion of these workers by Qt itself from the worker
      thread. As :class:`QRunnable` instances, workers provide *no* equivalent
      to the :attr:`QObject.destroyed` signal for deterministically responding
      to worker deletion. The closest equivalents are:

      * The custom :attr:`signals.finished` signal, which typically (but *not*
        necessarily) coincides with worker deletion.
      * The optional Python **finalizer** (i.e., special ``__del__()`` method)
        and corresponding optional ``callback`` parameter performing similar
        finalization accepted by the standard :func:`weakref.ref` function,
        which approximately but *not* exactly coincide with worker deletion and
        which impose unreasonable constraints on implementation -- notably, the
        inability to propagate or handle exceptions raised by these callables:

            Exceptions raised by the callback will be noted on the standard
            error output, but cannot be propagated; they are handled in exactly
            the same way as exceptions raised from an object’s ``__del__()``
            method.

        Exception propagation and logging is central to sane and deterministic
        application behaviour. Any small benefit gained from handling worker
        deletion via such a callback would surely be dwarfed by the grosser
        detriment of being unable to properly handle exceptions.

      In either case, there would exist a slice of time in which Qt has
      implicitly deleted a worker but Python has yet to be notified of that
      implicit deletion and hence continue to operate under the inappropriate
      assumption of that worker's continued existence. Race conditions ensue.
      Technically, these conditions could be obviated by:

      * Explicitly creating weak rather than strong references to workers.
      * Wrapping each access to such a weak reference in a
        ``try: ... except ReferenceError: ...`` block handling this race.

      In practice, however, the maintenance burden of doing so substantially
      exceeds the negligible benefit of permitting Qt to delete workers.

    Thread Affinity
    ----------
    All attributes of instances of this class reside in the original thread in
    which this worker was instantiated *except* the following, which reside in
    a dedicated thread of a parent :class:`QThreadPool` container and hence are
    guaranteed to be thread-safe by definition:

    * The :meth:`run` method.
    * All local objects instantiated by the :meth:`run` method.

    All other attributes should be assumed to *not* be thread-safe. These
    attributes may nonetheless be rendered thread-safe by either:

    * Locking access to these attributes behind Qt-specific mutual exclusion
      primitives and context managers (e.g., :class:`QMutexLocker`).
    * Defining these attributes to be Qt-specific atomic types (e.g.,
      :class:`QAtomicInt`). Since no Python-based Qt framework (including
      PySide2) exposes these types, this approach applies *only* to C++-based
      Qt applications. Ergo, atomic types are currently inapplicable. See below
      for further details.

    Signal-slot Connections
    ----------
    This class subclasses the :class:`QRunnable` interface rather than the
    standard :class:`QObject` base class. This has various real-world
    implications. In particular, subclasses of this class *cannot* directly
    participate in standard queued signal-slot connections and hence should
    define neither signals nor slots.

    Instead, to emit signals from this worker to slots on objects residing in
    other threads:

    * Declare a separate :class:`QObject` subclass defining these signals.
    * Instantiate an instance of this subclass in the parent thread in which
      this worker is instantiated.
    * Pass this instance to this worker's :meth:`__init__` method.

    By definition, this worker *cannot* receive any signals emitted from any
    objects residing in other threads as conventional slots. Instead:

    * Define each such slot as a simple method of this subclass. Since this
      method will be run from the other thread in which the object calling this
      method resides rather than the pooled thread in which this worker
      resides, this method's body *must* protectively guard access to instance
      variables via Qt-specific mutual exclusion primitives. While numerous
      primitives exist, the following maximize thread-safety in common edge
      cases (e.g., exceptions) and hence are strongly preferred:

      * :class:`QReadLocker` and :class:`QWriteLocker`, context managers
        suitable for general-purpose use in guarding access to variables
        safely:

        * Readable from multiple concurrent threads.
        * Writable from only a single thread at a time.

      * :class:`QMutexLocker`, a context manager suitable for general-purpose
        use in guarding access to variables safely readable *and* writable from
        only a single thread at a time.

    Lastly, note that Qt defines numerous atomic types publicly accessible to
    C++ but *not* Python applications (e.g., :class:`QtCore::QAtomicInt`). In
    theory, these types could be leveraged as an efficient alternative to the
    primitives listed above. In practice, these types have yet to be exposed
    via any Python Qt framework (PyQt5, PySide2, or otherwise) and hence remain
    a pipe dream at best.

    Versus QtConcurrent
    ----------
    The API published by this superclass bears a mildly passing resemblance to
    the API published by the QtConcurrent framework -- notably, the
    :class:`PySide2.QtCore.QFuture` class. Unfortunately, the latter imposes
    extreme constraints *not* imposed by this superclass.

    The QtConcurrent framework only provides a single means of multithreading
    arbitrary long-running tasks: :func:`PySide2.QtConcurrent.run`. The
    official documentation publicly admits the uselessness of this function:

        Note that the :class:`PySide2.QtCore.QFuture` returned by
        :func:`PySide2.QtConcurrent.run` does not support canceling, pausing,
        or progress reporting. The :class:`PySide2.QtCore.QFuture` returned can
        only be used to query for the running/finished status and the return
        value of the function.

    One enterprising StackOverflower `circumvented this constraint`_ by
    defining a robust C++ :class:`PySide2.QtCore.QFuture` analogue supporting
    canceling, pausing, and progress reporting. Sadly, this analogue requires
    C++-specific facilities unavailable under Python, including:

    * **Templating.** Since the :class:`PySide2.QtCore.QFuture` API is
      templated, all analogues of that API are also necessarily templated.
    * Private, undocumented Qt APIs (e.g., ``QFutureInterface``,
      ``QFutureInterfaceBase``).

    Ergo, the QtConcurrent framework is largely inapplicable in Python and
    certainly inapplicable for multithreading arbitrary long-running tasks.

    .. _circumvented this constraint:
       https://stackoverflow.com/a/16729619/2809027

    Versus QThread + QObject
    ----------
    The API published by this superclass also bears a passing resemblance to
    various third-party APIs duplicating the common worker-thread Qt model.
    These models typically:

    * Define one or more application-specific **worker types** (i.e.,
      :class:`QObject` subclasses performing long-running tasks to be
      multithreaded).
    * Instantiate these subclasses as local worker objects.
    * Instantiate a local :class:`QThread` object.
    * Move these workers into this thread via the :class:`QObject.moveToThread`
      method.
    * Start this thread by calling the :class:`QThread.start` method.
    * Start, pause, resume, cancel, and restart these workers thread by
      emitting signals connected to slots defined on these workers.

    As with the aforementioned QtConcurrent framework, this approach
    fundamentally works in C++ but fails in Python. For unknown reasons, the
    :class:`QObject.moveToThread` method silently fails to properly move worker
    objects in entirety from the main thread into the worker thread. Slots
    defined on workers claim to run from within their worker thread but instead
    run from within the main thread, as trivially observed with logging.

    Ergo, the worker-thread Qt model is also largely inapplicable in Python and
    certainly inapplicable for multithreading arbitrary long-running tasks.

    Attributes (Public)
    ----------
    signals : QBetseeThreadPoolWorkerSignals
        Low-level collection of all public signals thread-safely emittable by
        the :meth:`run` method from within an arbitrary pooled thread possibly
        running *no* Qt event loop.

    Attributes (Private)
    ----------
    _thread : WeakRefType
        Weak reference to the :class:`QThread` instance wrapping the thread in
        which the :meth:`run` method is currently running if that method is
        currently running *or* ``None`` otherwise (i.e., if this worker either
        has yet to be started *or* has already finished).
    _worker_id : int
        0-based integer uniquely identifying this worker. This worker is
        guaranteed to be the *only* instance of this class assigned this
        integer for the lifetime of the current process. For disambiguity with
        the :func:`id` builtin, this variable is *not* named ``_id``.

    Attributes (Private: State)
    ----------
    _state : ThreadWorkerState
        Non-thread-safe current execution state of this worker. This state is
        non-thread-safe and hence should *only* be accessed by instantiating an
        exception-safe :class:QMutexLocker` context manager nonce passed the
        :attr:`_state_lock` primitive as the target of a ``with`` context.
    _state_lock : QMutex
        Non-exception-safe mutual exclusion primitive rendering the
        :meth:`state` property thread-safe. This primitive is
        non-exception-safe and hence should *never* be accessed directly. Each
        access to this primitive should be encapsulated by instantiating an
        exception-safe :class:QMutexLocker` context manager nonce as the
        target of a ``with`` context. Note that the context provided by the
        :class:QMutexLocker` class is *not* safely reusable and hence *must* be
        re-instantiated in each ``with`` context.
    _state_unpaused : QWaitCondition
        Thread synchronization primitive, permitting this worker when paused in
        its parent thread to indefinitely block until an object in another
        thread (e.g., the main thread) requests this worker be unpaused by
        waking up this primitive and hence this worker.

    See Also
    ----------
    https://martinfitzpatrick.name/article/multithreading-pyqt-applications-with-qthreadpool
        Prominent blog article entitled "Multithreading PyQt applications with
        QThreadPool," strongly inspiring this implementation.
    https://stackoverflow.com/a/34302791/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self) -> None:
        '''
        Initialize this pooled worker.
        '''

        # Initialize our superclass.
        super().__init__()

        # Prevent Qt from implicitly deleting this worker on completion, as is
        # the C++-centric default. See the "Lifecycle" section of the above
        # docstring for further commentary.
        self.setAutoDelete(False)

        # 0-based integer uniquely identifying this worker.
        self._worker_id = _get_worker_id_next()

        # Mutual exclusion objects safeguarding this worker.
        self._state_lock = QMutex()
        self._state_unpaused = QWaitCondition()

        # Default this worker's initial state to the idle state.
        self._state = ThreadWorkerState.IDLE

        # Weak reference to the thread currently running the run() method.
        self._thread = None

        # Collection of all public signals emittable by this worker, classified
        # *AFTER* all other instance variables above to enable subclass methods
        # to safely reference these variables.
        self.signals = self._make_signals()


    @type_check
    def init(
        self,
        progress_bar: QBetseeProgressBarOrNoneTypes = None,
        progress_label: QLabelOrNoneTypes = None,
        handler_failed: CallableOrNoneTypes = None,
        handler_finished: CallableOrNoneTypes = None,
    ) -> None:
        '''
        Finalize this pooled worker's initialization with the passed widgets.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        progress_bar : QBetseeProgressBarOrNoneTypes
            Progress bar to connect to numeric progress signals emitted by this
            worker. Defaults to ``None``, in which case the caller is expected
            to manually connect these signals to appropriate widget slots.
        progress_label : QLabelOrNoneTypes
            Label to connect to the
            :attr:`QBetseeThreadPoolWorkerSignals.progress_stated` signal
            emitted by this worker. Defaults to ``None``, in which case the
            caller is expected to manually connect this signal to appropriate
            widget slots.
        handler_failed : CallableOrNoneTypes
            Slot to connect to the
            :attr:`QBetseeThreadPoolWorkerSignals.failed` signal emitted by
            this worker. Defaults to ``None``, in which case the caller is
            expected to manually connect this signal to appropriate slots.
        handler_finished : CallableOrNoneTypes
            Slot to connect to the
            :attr:`QBetseeThreadPoolWorkerSignals.finished` signal emitted by
            this worker. Defaults to ``None``, in which case the caller is
            expected to manually connect this signal to appropriate slots.
        '''

        # If passed a progress bar, connect progress signals emitted by this
        # worker to the corresponding slots of this progress bar.
        if progress_bar is not None:
            # Connect the ranged signal to the custom set_range_and_value_min()
            # slot rather than the stock setRange() slot, which fails to set
            # this progress bar's value to the minimum value of this range.
            self.signals.progress_ranged.connect(
                progress_bar.set_range_and_value_minimum)
            self.signals.progressed.connect(progress_bar.setValue)

        # If passed a label, connect progress signals emitted by this worker to
        # the corresponding slots of this label.
        if progress_label is not None:
            self.signals.progress_stated.connect(progress_label.setText)

        # For each passed slot, connect the corresponding signals emitted by
        # this worker to that slot.
        if handler_failed is not None:
            self.signals.failed.connect(handler_failed)
        if handler_finished is not None:
            self.signals.finished.connect(handler_finished)

    # ..................{ PROPERTIES                        }..................
    @property
    def _is_running(self) -> bool:
        '''
        ``True`` only if this worker is **running** (i.e., performing
        subclass-specific business logic, typically long-running).

        Caveats
        ----------
        **This private property is non-thread-safe.** The caller *must*
        explicitly embed each access of this property within a context manager
        of the form ``with QMutexLocker(self._state_lock):``.
        '''

        return self._state is ThreadWorkerState.RUNNING

    # ..................{ SLOTS                             }..................
    #FIXME: It would be great to additionally set the object name of (and hence
    #the name of the process associated with) the parent thread of this worker
    #to a human-readable name based on the subclass of this worker. For similar
    #logic, see the QBetseeObjectMixin.set_obj_name_from_class_name() method.
    def run(self):
        '''
        Thread-safe psuedo-slot (i.e., non-slot method mimicking the
        thread-safe, push-based action of a genuine method) performing *all*
        subclass-specific business logic for this worker.

        This method works in a thread-safe manner safely pausable, resumable,
        and stoppable at any time from any object in any thread by directly
        calling the equally thread-safe :meth:`pause`, :meth:`resume`, and
        :meth :meth:`stop` methods.

        States
        ----------
        If this worker is in the :attr:`ThreadWorkerState.IDLE` state, this
        method changes to the :attr:`ThreadWorkerState.RUNNING` state and calls
        the subclass :meth:`_work` method.

        If this worker is in the :attr:`ThreadWorkerState.PAUSED` state, this
        method interprets this signal as a request to resume the work
        presumably previously performed by this worker by a prior signalling of
        this method.  To avoid reentrancy issues, this method changes to the
        :attr:`ThreadWorkerState.RUNNING` state and immediately returns.
        Assuming that a prior call to this method is still executing, that call
        will internally detect this change and resume working as expected.

        If this worker is in the :attr:`ThreadWorkerState.RUNNING` state, this
        method interprets this signal as an accidental attempt by an external
        caller to re-perform the work concurrently being performed by a prior
        call to this method. In that case, this method safely logs a non-fatal
        warning and immediately returns.

        See the :meth:`pause` method for commentary on these design decisions.

        Signals
        ----------
        This method emits the following signals (in order):

        * :attr:`signals.started` immediately *before* this method performs any
          subclass work.
        * :attr:`signals.failed` immediately *after* this method erroneously
          raises an unexpected exception while performing subclass work but
          *before* the :attr:`signals.finished` signal is emitted.
        * :attr:`signals.succeeded` immediately *after* all subclass work
          performed by this method successfully returns but *before* the
          :attr:`signals.finished` signal is emitted.
        * :attr:`signals.finished` immediately *after* this method performs all
          subclass work, regardless of whether that work succeeded or failed.

        Caveats
        ----------
        This method is *not* intended to be called directly. Only the parent
        thread pool of this worker is intended to call this method. External
        callers are advised to call the :func:`guipoolthread.start_worker`
        function instead.

        Subclasses should override the :meth:`_work` method rather than this
        method to perform subclass-specific business logic. This method is
        neither intended nor designed to be redefined by subclasses.
        '''

        # True only if the _work() method called below returns successfully
        # (i.e., *WITHOUT* raising exceptions). Defaults to False for safety.
        is_success = False

        # To ensure that callers receive notification of *ALL* exceptions
        # raised by this method, the remainder of this method *MUST* be
        # embedded within an exception handler.
        try:
            # Log this run.
            guithread.log_debug_thread_current(
                'Starting pooled thread worker "%d"...', self._worker_id)

            # If this thread is the main thread rather than a pooled thread,
            # raise an exception. Allowing the main thread to directly call
            # this method could, in the worst case, result in that thread and
            # thus the GUI event loop being indefinitely blocked (e.g., by a
            # subsequent call to the pause() pseudo-slot).
            guithread.die_if_thread_current_main()

            # Classify the pooled thread running this worker *BEFORE*
            # performing any work, which assumes this thread to exist.
            self._thread = pyref.refer_weak(guithread.get_thread_current())

            # Within a thread- and exception-safe context manager synchronizing
            # access to this state across multiple threads...
            with QMutexLocker(self._state_lock):
                # If this worker is *NOT* currently idle (i.e., is either
                # currently working or paused), the Universe is on the verge of
                # imploding. In short, this should *NEVER* happen; if it does,
                # the caller probably erroneously attempted to manually call
                # this method multiple times from the calling thread.
                #
                # Regardless, this constitutes an error. Raise an exception!
                if self._state is not ThreadWorkerState.IDLE:
                    raise BetseePySideThreadWorkerException(
                        'Non-reentrant thread "{}" worker "{}" run() method '
                        'called reentrantly (i.e., multiple times).'.format(
                            guithread.get_thread_current_id(),
                            self._worker_id))

                # Change to the running state.
                self._state = ThreadWorkerState.RUNNING

            # Notify external subscribers *BEFORE* beginning subclass work.
            self.signals.started.emit()

            # If this worker or this worker's thread has been externally
            # requested to halt immediately after being requested to start,
            # respect this wish. While an unlikely edge case, the fixed cost of
            # this test is negligible. Ergo, we do so.
            self._halt_work_if_requested()

            # Retain this purely for exception testing purposes.
            # raise ValueError('wat?')

            # Value returned by performing subclass-specific business logic.
            return_value = self._work()
        # If a periodic call to the _halt_work_if_requested() method performed
        # within the above call detects either this worker or this worker's
        # thread has been externally requested to stop, do so gracefully by...
        except BetseePySideThreadWorkerStopException:
            # Log this cessation. That's it. Welp, that was easy.
            guithread.log_debug_thread_current(
                'Stopping pooled thread worker "%d"...', self._worker_id)
        # If this worker raised any other exception...
        except Exception as exception:
            # Log this failure.
            guithread.log_debug_thread_current(
                'Emitting pooled thread worker "%d" exception "%r"...',
                self._worker_id, exception)

            # Emit this exception to external subscribers.
            self.signals.failed.emit(exception)
        # Else, this worker raised no exception. In this case...
        else:
            # Note the _work() method called above to have returned
            # successfully (i.e., *WITHOUT* raising exceptions).
            is_success = True

            # Log this success.
            guithread.log_debug_thread_current(
                'Returning pooled thread worker "%d" value "%r"...',
                self._worker_id, return_value)

            # Emit this return value to external subscribers.
            self.signals.succeeded.emit(return_value)
        # In either case, this worker completed. In this case...
        finally:
            # Log this completion.
            guithread.log_debug_thread_current(
                'Finishing pooled thread worker "%d" with exit status "%r"...',
                self._worker_id, is_success)

            # Within a thread- and exception-safe context manager synchronizing
            # access to this state across multiple threads...
            with QMutexLocker(self._state_lock):
                # If this worker is running, set this worker's state to the
                # idle (i.e., non-running) state.
                if self._is_running:
                    self._state = ThreadWorkerState.IDLE

            # Emit this completion status to external subscribers.
            self.signals.finished.emit(is_success)

            # Declassify the pooled thread running this worker *AFTER*
            # performing all work, which assumes this thread to exist.
            self._thread = None

    # ..................{ SLOTS ~ pause                     }..................
    def pause(self) -> None:
        '''
        Thread-safe psuedo-slot (i.e., non-slot method mimicking the
        thread-safe push-based action of a genuine slot) pausing all work
        performed by this worker.

        This slot temporarily halts this work in a thread-safe manner safely
        resumable at any time by calling the :meth:`resume` method.

        States
        ----------
        If this worker is in the :attr:`ThreadWorkerState.RUNNING` state,
        this method pauses work by changing this state to
        :attr:`ThreadWorkerState.PAUSED`. The :meth:`_halt_work_if_requested`
        method in the thread running this worker then detects this state change
        and responds by indefinitely blocking on this synchronization primitive
        until subsequently awoken by a call to the :meth:`resume` method.

        If this worker is in any other state, this method silently reduces to a
        noop and hence preserves the existing state.
        '''

        # Log this action.
        guithread.log_debug_thread_current(
            'Requesting pooled thread worker "%d" to pause...',
            self._worker_id)

        # Within a thread- and exception-safe context manager synchronizing
        # access to this state across multiple threads...
        with QMutexLocker(self._state_lock):
            # If this worker is *NOT* currently working...
            if self._state is not ThreadWorkerState.RUNNING:
                # Log this attempt.
                guithread.log_debug_thread_current(
                    'Ignoring attempt to pause idle or already paused '
                    'pooled thread worker "%d".', self._worker_id)

                # Safely reduce to a noop.
                return
            # Else, this worker is currently working.

            # Change this worker's state to paused.
            self._state = ThreadWorkerState.PAUSED


    def resume(self) -> None:
        '''
        Thread-safe psuedo-slot (i.e., non-slot method mimicking the
        thread-safe push-based action of a genuine slot) unpausing this worker.

        This method resumes work in a thread-safe manner safely re-pausable at
        any time by re-calling the :meth:`pause` method.

        States
        ----------
        If this worker is in the :attr:`ThreadWorkerState.PAUSED` state,
        this method resumes work by changing this state to
        :attr:`ThreadWorkerState.RUNNING` and waking up the currently blocked
        synchronization primitive if any. The :meth:`_halt_work_if_requested`
        method in the thread running this worker then detects this state change
        and responds by unblocking from this primitive and returning to the
        subclass-specific :meth:`_work` method.

        If this worker is in any other state, this method silently reduces to a
        noop and hence preserves the existing state.
        '''

        # Log this action.
        guithread.log_debug_thread_current(
            'Requesting pooled thread worker "%d" to resume...',
            self._worker_id)

        # Within a thread- and exception-safe context manager synchronizing
        # access to this state across multiple threads...
        with QMutexLocker(self._state_lock):
            # Attempt to...
            try:
                # If this worker is *NOT* currently paused...
                if self._state is not ThreadWorkerState.PAUSED:
                    # Log this attempt.
                    guithread.log_debug_thread_current(
                        'Ignoring attempt to resume idle or already working '
                        'pooled thread worker "%d".', self._worker_id)

                    # Safely reduce to a noop.
                    return
                # Else, this worker is currently paused.

                # Change this worker's state to working, thus unpausing.
                self._state = ThreadWorkerState.RUNNING
            # Regardless of whether doing so raised an exception or not...
            finally:
                # Unblock the parent thread of this worker if currently
                # blocked. Deadlocks are unlikely but feasible in edge cases
                # unless this method is unconditionally called here.
                self._unblock_work()

    # ..................{ PSEUDO-SLOTS ~ stop               }..................
    def stop(self) -> None:
        '''
        Thread-safe psuedo-slot (i.e., non-slot method mimicking the
        thread-safe push-based action of a genuine slot) gracefully and
        permanently halting all work performed by this worker.

        As a necessary side effect, this pseudo-slot also resumes this worker
        if currently paused *before* halting this worker, thus unblocking the
        parent thread of this worker if currently blocked.

        Caveats
        ----------
        By :class:`QRunnable` design, this worker is *not* safely restartable
        after halting -- whether by this method being called, the :meth:`_work`
        method either raising an exception or returning successfully without
        doing so, the parent thread running this worker being terminated, or
        otherwise. Ergo, completion is permanent.

        States
        ----------
        Regardless of the current state of this worker, this method halts work
        by changing this state to :attr:`ThreadWorkerState.IDLE`. The
        :meth:`_halt_work_if_requested` method in the thread running this
        worker then detects this state change and responds by raising an
        exception internally caught by the parent :meth:`run` method.
        '''

        # Log this action.
        guithread.log_debug_thread_current(
            'Requesting pooled thread worker "%d" to stop...', self._worker_id)

        # Within a thread- and exception-safe context manager synchronizing
        # access to this state across multiple threads...
        with QMutexLocker(self._state_lock):
            # Attempt to...
            try:
                # Regardless of the current state of this worker, change this
                # worker's state to idle (i.e., non-working).
                self._state = ThreadWorkerState.IDLE
            # Regardless of whether doing so raised an exception or not...
            finally:
                # Unblock the parent thread of this worker if currently
                # blocked. Deadlocks are unlikely but possible in edge cases
                # unless this method is unconditionally called here.
                self._unblock_work()


    def delete_later(self) -> None:
        '''
        Thread-safe psuedo-slot (i.e., non-slot method mimicking the
        thread-safe push-based action of a genuine slot) gracefully stopping
        this worker if needed *and* scheduling all :class:`QObject` instances
        owned by this worker for immediate deletion.

        Specifically, this method schedules all signals owned by this worker
        for immediate deletion and hence disconnection from all slots they are
        currently connected to. (Note that doing so is technically unnecessary
        in most cases. In theory, the subsequent nullification of this worker
        by the call should suffice to schedule this deletion. As doing so has
        no harmful side effects, however, this method exists.)
        '''

        # Log this action.
        guithread.log_debug_thread_current(
            'Scheduling pooled thread worker "%d" for deletion...',
            self._worker_id)

        # Within a thread- and exception-safe context manager synchronizing
        # access to this state across multiple threads...
        with QMutexLocker(self._state_lock):
            # If this worker is running, gracefully stop this worker and hence
            # emit both the "stopped" and "finished" signals *BEFORE*
            # scheduling these signals for deletion.
            if self._is_running:
                self.stop()

            # If this worker's signals are still alive...
            if self.signals is not None:
                # Schedule these signals for deletion.
                self.signals.deleteLater()

                # Nullify these signals for safety.
                self.signals = None

            # Change this worker's state to deleted *AFTER* successfully
            # scheduling all "QObject" instances owned by this worker.
            self._state = ThreadWorkerState.DELETED


    def halt(self) -> None:
        '''
        Thread-safe psuedo-slot (i.e., non-slot method mimicking the
        thread-safe push-based action of a genuine slot) attempting to first
        gracefully and then non-gracefully terminate both this worker and the
        pooled thread running this worker... **by any means necessary.**

        Specifically, this method (in order):

        #. Politely requests this pooled thread to gracefully terminate.
        #. Blocks the calling thread (e.g., the main thread) for a negligible
           span of time (e.g., 100ms) while waiting for this pooled thread to
           gracefully terminate. Since the :func:`guipoolthread.halt_workers`
           function calling this pseudo-slot already blocked for a lengthy span
           of time without success for this worker to gracefully terminate,
           this pseudo-slot effectively performs *no* blocking.
        #. If this pooled thread is still running and hence failed to
           gracefully terminate, this thread is immediately terminated
           non-gracefully. (Note that doing so may induce catastrophic damage.
           See the "Caveats" below.)

        Caveats
        ----------
        This pseudo-slot may induce data loss or corruption in edge cases in
        which this worker fails to gracefully terminate itself within a
        reasonable span of time. This implies that this method should *only* be
        called when absolutely necessary (e.g., at application shutdown).
        '''

        # Maximum number of milliseconds to block the current thread waiting
        # for the currently working worker (if any) to gracefully halt.
        WAIT_MAX_MILLISECONDS = 100  # 100ms

        # Log this action.
        guithread.log_debug_thread_current(
            'Terminating pooled thread worker "%d" shortly...',
            self._worker_id)

        # Within a thread- and exception-safe context manager synchronizing
        # access to this state across multiple threads...
        with QMutexLocker(self._state_lock):
            # If this worker is running, try to gracefully halt this worker
            # *BEFORE* coercively and hence non-gracefully halting this worker.
            if self._is_running:
                self.stop()

        # Strong reference to the thread still running this worker if any
        # or "None" otherwise. Note that this usage pattern guarantees this
        # thread to remain alive for the duration of this method call, as
        # strongly (...get it?) recommended by official documentation:
        #     https://docs.python.org/3/library/weakref.html#weak-reference-objects
        #
        # To avoid blocking attempts to gracefully halt this worker and hence
        # thread from other threads, the following logic is intentionally *NOT*
        # nested within the above context manager.
        thread = self._thread()

        # If this thread is no longer running this worker, log this fact
        # and return immediately.
        if thread is None:
            guithread.log_debug_thread_current(
                'Terminated pooled thread worker "%d" gracefully...',
                self._worker_id)
        # Else, this thread is still running this worker. In this case...
        else:
            # Log this fact.
            guithread.log_debug_thread_current(
                'Terminating pooled thread worker "%d" '
                'possibly non-gracefully...',
                self._worker_id)

            # Politely request this thread to gracefully terminate. Since
            # the prior call to the stop() pseudo-slot failed to do so,
            # this call is unlikely to fare any better. Since performing
            # this call incurs no cost, there's no loss in trying anyway.
            thread.quit()

            # Block the calling thread (e.g., main thread) for a negligible
            # window of time in the desperate hope that this working thread
            # will gracefully terminate within this window.
            #
            # Since the guipoolthread.halt_workers() function calling this
            # slot has already waited for 30 seconds without success for this
            # worker to gracefully terminate, avoid blocking for a significant
            # window of time here.
            thread.wait(WAIT_MAX_MILLISECONDS) # 100ms

            # If the working thread is still working, we have little choice
            # but to...
            if thread.isRunning():
                # Log this fact.
                guithread.log_warning_thread_current(
                    'Terminating pooled thread worker "%d" '
                    'non-gracefully!', self._worker_id)

                # Terminate this thread non-gracefully. *UGH!*
                thread.terminate()
            # Else, the working thread is no longer working. Log this fact
            # and return immediately.
            else:
                guithread.log_debug_thread_current(
                    'Terminating pooled thread worker "%d" '
                    'gracefully after all...', self._worker_id)

    # ..................{ MAKERS ~ optional                 }..................
    # Concrete methods designed to be safely redefinable by subclasses.

    def _make_signals(self) -> QBetseeThreadPoolWorkerSignals:
        '''
        Low-level collection of all public signals thread-safely emittable by
        the :meth:`run` method from within an arbitrary pooled thread possibly
        running *no* Qt event loop.

        Design
        ----------
        Subclasses may optionally expose subclass-specific signals by trivially
        redefining this method to return a subclass-specific
        :class:`QBetseeThreadPoolWorkerSignals` instance.
        '''

        return QBetseeThreadPoolWorkerSignals(
            # Weak reference to the _halt_work_if_requested() method, avoiding
            # circular references between this object and this child object.
            halt_work_if_requested=pyref.refer_weak(
                self._halt_work_if_requested))

    # ..................{ WORKERS ~ abstract                }..................
    # Abstract methods required to be redefined by subclasses.

    def _work(self) -> None:
        '''
        Perform *all* subclass-specific business logic for this worker.

        The superclass :meth:`start` slot internally calls this method in a
        thread-safe manner safely pausable *and* stoppable at any time (e.g.,
        by emitting a signal connected to the :meth:`pause` or :meth:`stop`
        slots).

        Design
        ----------
        Subclasses are required to redefine this method to perform this logic
        in an iterative manner periodically calling the
        :meth:`_halt_work_if_requested` method.

        If either:

        * This worker has been externally signalled to stop (e.g., by emitting
          a signal connected to the :meth:`stop` slot).
        * The thread of execution currently running this worker has been
          externally requested to stop (e.g., by calling the
          :func:`guithread.halt_thread_work` function).

        Then the next such call to the :meth:`_halt_work_if_requested` method
        will raise an exception caught by the parent :meth:`start` slot,
        signalling that slot to immediately terminate this worker. Ergo, that
        method should be called *only* when the subclass is in an
        **interruptible state** (i.e., a self-consistent internal state in
        which this subclass is fully prepared to be immediately terminated).
        '''

        # The next best thing to a properly abstract method, given "QObject"
        # constraints against declaring an "ABCMeta" metaclass. *shrug*
        raise BetseMethodUnimplementedException()

    # ..................{ WORKERS ~ concrete                }..................
    # Concrete methods intended to be called but *NOT* overriden by subclasses.

    def _halt_work_if_requested(self) -> None:
        '''
        Thread-safely either temporarily or permanently halt all
        subclass-specific business logic when requested to do so by external
        callers residing in other threads.

        This function is intended to be periodically called by the subclass
        :meth:`_work` function. Specifically, this method:

        * Raises an exception (expected to be caught by the :meth:`run` method
          calling the :meth:`_work` function) if either:

          * This worker has been externally signalled to stop (e.g., by an
            external call to the the :meth:`stop` method).
          * The thread of execution currently running this worker has been
            externally requested to stop (e.g., by calling the
            :func:`guithread.halt_thread_work` function).

        * Pauses this worker if this worker has been externally signalled to
          pause (e.g., by an external call to the the :meth:`pause` method).

        Caveats
        ----------
        This method imposes minor computational overhead and hence should be
        called intermittently (rather than overly frequently). Notably, each
        call to this method processes *all* pending events currently queued
        with this worker's thread -- including those queued for all other
        workers currently running in this thread.

        Raises
        ----------
        BetseePySideThreadWorkerStopException
            If this worker or this worker's thread of execution has been
            signalled or requested to be stopped.
        '''

        # Within a thread- and exception-safe context manager synchronizing
        # access to this state across multiple threads...
        with QMutexLocker(self._state_lock):
            # If either:
            if (
                # If an external call to the stop() method has requested that
                # this worker be stopped...
                self._state is ThreadWorkerState.IDLE or
                # This worker's thread has been externally requested to stop...
                guithread.should_halt_thread_work()
            # ...then permanently halt this worker.
            ):
                self._stop_work()

            # If an external call to the pause() method from another thread has
            # requested this worker to temporarily halt work, do so *AFTER*
            # detecting and handling a request to permanently halt work. Why?
            # Requests to permanently halt take priority over requests to
            # temporarily halt.
            if self._state is ThreadWorkerState.PAUSED:
                self._block_work()

    # ..................{ STOPPERS                          }..................
    def _stop_work(self) -> None:
        '''
        Thread-safely gracefully terminate all subclass-specific business logic
        performed by this worker.

        This method raises an internal exception expected to be caught *only*
        by the parent :meth:`run` method, as a crude form of coordinating
        signalling between this and that method.
        '''

        # Log this action.
        guithread.log_debug_thread_current(
            'Stopping pooled thread worker "%d"...', self._worker_id)

        # Instruct the parent run() method to stop.
        raise BetseePySideThreadWorkerStopException('So say we all.')

    # ..................{ (UN)BLOCKERS                      }..................
    def _block_work(self) -> None:
        '''
        Non-thread-safely indefinitely block all subclass-specific business
        logic performed by this worker.

        This method waits for an external call from another thread to a worker
        pseudo-slot (e.g., the :meth:`resume` method), all of which internally
        call the :meth:`_unblock_work` method to safely resume work.

        Caveats
        ----------
        **This private method is non-thread-safe.** The caller *must*
        explicitly embed each call to this method within a context manager of
        the form ``with QMutexLocker(self._state_lock):``.
        '''

        # Log this blocking behaviour.
        guithread.log_debug_thread_current(
            'Pausing pooled thread worker "%d"...', self._worker_id)

        # Notify callers that this worker is now paused immediately *BEFORE*
        # blocking this worker.
        self.signals.paused.emit()

        # Indefinitely wait for an external call to the resume() method from
        # another thread to request this worker to resume work. This has
        # several consequences, including (in temporal order):
        #
        # 1. The "_state_lock" *MUST* be acquired (i.e., locked) before calling
        #    the QWaitCondition.wait() method on this lock. This is the
        #    caller's responsibility, sadly.
        # 2. The QWaitCondition.wait() method immediately releases (i.e.,
        #    unlocks) this lock.
        # 3. The QWaitCondition.wait() method blocks indefinitely until the
        #    QWaitCondition.wakeOne() or wakeAll() methods are called -- in
        #    this case, by a call to our resume() method.
        # 4. The QWaitCondition.wait() method immediately re-acquires (i.e.,
        #    relocks) this lock.
        # 5. The QWaitCondition.wait() method returns.
        #
        # While it is typically inadvisable to mix "QMutexLocker"-style
        # implicit lock management with "QMutex"-style explicit lock control,
        # doing so here is the simplest, sanest, and safest approach.
        self._state_unpaused.wait(self._state_lock)

        # Notify callers that this worker is now resumed immediately *AFTER*
        # unblocking this worker.
        self.signals.resumed.emit()


    def _unblock_work(self) -> None:
        '''
        Non-thread-safely unblock all subclass-specific business logic
        performed by this worker.

        This method wakes up the parent thread of this worker and hence this
        worker *after* the :meth:`_halt_work_if_requested` previously called
        the :meth:`_block_work` method to indefinitely block that thread.
        Moreover, this method is typically called by an external call in
        another thread to a worker pseudo-slot (e.g., the :meth:`resume`
        method), all of which internally call this method to request that this
        worker safely resume work.

        If the parent thread of this worker is *NOT* currently blocked, this
        method silently reduces to a noop.

        Caveats
        ----------
        **This private method is non-thread-safe.** The caller *must*
        explicitly embed each call to this method within a context manager of
        the form ``with QMutexLocker(self._state_lock):``.
        '''

        # Log this unblocking behaviour.
        guithread.log_debug_thread_current(
            'Resuming pooled thread worker "%d"...', self._worker_id)

        # Wake up the parent thread of this worker if currently blocking.
        self._state_unpaused.wakeOne()

# ....................{ SUPERCLASSES ~ callable           }....................
class QBetseeThreadPoolWorkerCallable(QBetseeThreadPoolWorker):
    '''
    Low-level **callable-defined pooled worker** (i.e., pooled worker whose
    business logic is encapsulated by a caller-defined callable at worker
    initialization time).

    This superclass is a convenience wrapper for the
    :class:`QBetseeThreadPoolWorker` superclass, simplifying usage in the
    common case of business logic definable by a single callable. Under the
    more general-purpose :class:`QBetseeThreadPoolWorker` API, each novel type
    of business logic must be implemented as a distinct subclass overriding the
    :meth:`_work` method to perform that specific logic. Under the less
    general-purpose API provided by this superclass, each such type of business
    logic is instead implemented as a simple callable (e.g., function, lambda)
    performing that specific logic; no new subclasses need be defined.

    Attributes
    ----------
    _func : CallableTypes
        Callable to be subsequently called by the :meth:`start` method,
        performing all business logic isolated to this worker within its parent
        thread.
    _func_args : SequenceTypes
        Sequence of all positional arguments to be passed to the :func:`func`
        callable when subsequently called.
    _func_kwargs : MappingType
        Mapping of all keyword arguments to be passed to the :func:`func`
        callable when subsequently called.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        func: CallableTypes,

        # Optional parameters.
        func_args: SequenceOrNoneTypes,
        func_kwargs: MappingOrNoneTypes,
    ) -> None:
        '''
        Initialize this callable-defined pooled worker with the passed callable
        and positional and keyword arguments to be passed to that callable when
        subsequently called by the :meth:`start` method.

        Parameters
        ----------
        func : CallableTypes
            Callable to be subsequently called by the :meth:`start` method,
            performing all business logic isolated to this worker within its
            parent thread.
        func_args : SequenceOrNoneTypes
            Sequence of all positional arguments to be passed to the
            :func:`func` callable when subsequently called. Defaults to
            ``None``, in which case this sequence defaults to the empty tuple.
        func_kwargs : MappingOrNoneTypes
            Mapping of all keyword arguments to be passed to the :func:`func`
            callable when subsequently called. Defaults to ``None``, in which
            case this mapping defaults to the empty dictionary.
        '''

        # Initialize our superclass.
        super().__init__()

        # Default all unpassed optional parameters to sane defaults.
        if func_args is None:
            func_args = ()
        if func_kwargs is None:
            func_kwargs = {}

        # Classify all passed parameters *AFTER* defaulting these parameters.
        self._func = func
        self._func_args = func_args
        self._func_kwargs = func_kwargs

    # ..................{ WORKERS                           }..................
    def _work(self) -> object:

        # Call this callable with these positional and keyword arguments and
        # return the value returned by this callable, which the parent run()
        # method will emit to slots connected to the "succeeded" signal.
        return self._func(*self._func_args, **self._func_kwargs)

# ....................{ GETTERS                           }....................
def _get_worker_id_next() -> int:
    '''
    Thread-safe 0-based integer uniquely identifying the next **pooled worker**
    (i.e., instance of the :class:`QBetseeThreadPoolWorker` superclass).

    This function internally increments this integer in a thread-safe manner,
    ensuring each of several concurrently instantiated workers to be assigned a
    unique 0-based integer.
    '''

    # Permit this global to be thread-safely incremented.
    global _worker_id_next

    # Within a thread- and exception-safe context manager synchronizing access
    # to this global across multiple threads...
    with QMutexLocker(_worker_id_lock):
        # 0-based integer uniquely identifying the next pooled worker.
        worker_id_next = _worker_id_next

        # Thread-safely increment this integer.
        _worker_id_next += 1

        # Return this integer.
        return worker_id_next



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based object encapsulating simulation configuration state.
'''

#FIXME: Currently, entering either <Ctrl-z> or <Ctrl-Shift-z> while in an
#editable widget that takes the keyboard focus forwards these key sequence
#events to that specific widget rather than the application as a whole,
#preventing the global undo stack from being modified by keyboard. To address
#this, consider carefully:
#
#* Redefining the key sequence event handler for each editable simulation
#  configuration widget as follows:
#  * If widget.isUndoRedoEnabled() returns True, then defer to the default
#    superclass key sequence event handler for this widget. This preserves the
#    existing functionality for widgets with pending actions to be undone.
#  * Else (i.e., if this widget has no pending actions to be undone), forward
#    this key sequence event to the global application or main window widgets,
#    triggering an undo or redo from the global undo stack.
#
#For further details on installing a custom eventFilter() specifically handling
#<Ctrl-z> keyboard shortcuts from editable widgets, see the following answer:
#    https://stackoverflow.com/a/28867475/2809027

#FIXME: Whenever the undo stack returns to the clean state through undoing and
#redoing commands, it emits the signal cleanChanged(). This signal should be
#connected to a slot simply emitting our more general-purpose
#"set_dirty_signal". Hmmmm; wait. Actually, is doing so even necessary? Since
#each application of an undo or redo operation necessarily invokes an editable
#widget slot already emitting "set_dirty_signal", there appears to be no need
#for this undo stack to explicitly do so as well.

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, QSize  # Signal, Slot
from PySide2.QtGui import QIcon, QKeySequence
from PySide2.QtWidgets import QUndoCommand, QUndoStack
from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideMenuException
from betsee.gui.window.guimainwindow import QBetseeMainWindow
from betsee.gui.simconf.guisimconf import QBetseeSimConf

# ....................{ CLASSES                            }....................
class QBetseeUndoStackSimConf(QUndoStack):
    '''
    :class:`QUndoStack`-based stack of all :class:`QBetseeUndoCommandSimConf`
    instances signifying user-driven simulation configuration modifications and
    the capacity to undo those modifications.

    Attributes (Non-widgets)
    ----------
    _redo_action : QAction
        Redo action synchronized with the contents of this stack.
    _undo_action : QAction
        Undo action synchronized with the contents of this stack.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        main_window: QBetseeMainWindow,
        sim_config: QBetseeSimConf,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this object, owned by the passed objects.

        To avoid circular references, this method is guaranteed to *not* retain
        a reference to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        sim_config: QBetseeSimConf
            Direct parent object against which to initialize this object.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Initialize all instance variables for safety.
        self._redo_action = None
        self._undo_action = None

        # Log this initialization.
        logs.log_debug('Customizing simulation configuration undo stack...')

        # Create all actions and icons associated with this undo stack.
        self._init_actions()

        # Create all items of the "Edit" menu requiring these actions.
        self._init_menu_edit(main_window)

        # Create all buttons of the main toolbar requiring these actions.
        self._init_toolbar(main_window)

        #FIXME: Connect these actions to appropriate "sim_config" slots and
        #signals. (See the "FIXME" above for commentary on exactly what.)


    @type_check
    def _init_actions(
        self,
        # main_window: QBetseeMainWindow,
        # sim_config: QBetseeSimConf,
    ) -> None:
        '''
        Create all actions and icons associated with this undo stack.

        Design
        ----------
        To synchronize the state and text of these actions with the contents of
        this undo stack, this method creates these actions by calling the
        :meth:`createUndoAction` and :meth:`createRedoAction` methods. Since Qt
        Designer lacks support for doing so, this method does so manually.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget.
        sim_config: QBetseeSimConf
            Direct parent object against which to initialize this object.
        '''

        # Redo icon associated with this redo action.
        redo_icon = QIcon()
        redo_icon.addFile(
            ':/icon/open_iconic/action-redo.svg',
            QSize(), QIcon.Normal, QIcon.Off,)

        # Undo icon associated with this undo action.
        undo_icon = QIcon()
        undo_icon.addFile(
            ':/icon/open_iconic/action-undo.svg',
            QSize(), QIcon.Normal, QIcon.Off,)

        # Redo action synchronized with the contents of this stack.
        self._redo_action = self.createRedoAction(
            self, QCoreApplication.translate(
                'QBetseeUndoStackSimConf', '&Redo'))
        self._redo_action.setIcon(redo_icon)
        self._redo_action.setObjectName('action_redo')
        self._redo_action.setShortcuts(QKeySequence.Redo)

        # Undo action synchronized with the contents of this stack.
        self._undo_action = self.createUndoAction(
            self, QCoreApplication.translate(
                'QBetseeUndoStackSimConf', '&Undo'))
        self._undo_action.setIcon(undo_icon)
        self._undo_action.setObjectName('action_undo')
        self._undo_action.setShortcuts(QKeySequence.Undo)


    @type_check
    def _init_menu_edit(self, main_window: QBetseeMainWindow) -> None:
        '''
        Create all items of the ``Edit`` menu requiring actions previously
        created by the :meth:`_init_actions` method.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget.
        '''

        # List of all actions in the "Edit" menu.
        menu_edit_actions = main_window.menu_edit.actions()

        # First action in the "Edit" menu, guaranteed by the logic below to be
        # separator. As a fallback in the event this menu contains no actions,
        # this action intentionally defaults to a placeholder integer
        # instructing subsequent insertions to append to this menu. While this
        # edge case should never happen, never is the enemy of reason.
        first_separator = 0

        # If at least one such action exists...
        if menu_edit_actions:
            # First such action.
            first_separator = menu_edit_actions[0]

            # If this action is *NOT* a separator, raise an exception.
            if not first_separator.isSeparator():
                raise BetseePySideMenuException(QCoreApplication.translate(
                    'QBetseeUndoStackSimConf',
                    'First "Edit" menu action '
                    '"{0}" not a separator.'.format(first_separator.text())))

        # Insert undo and redo actions before this separator in this menu.
        main_window.menu_edit.insertAction(first_separator, self._undo_action)
        main_window.menu_edit.insertAction(first_separator, self._redo_action)


    @type_check
    def _init_toolbar(self, main_window: QBetseeMainWindow) -> None:
        '''
        Create all buttons of the main toolbar requiring actions previously
        created by the :meth:`_init_actions` method.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget.
        '''

        # List of all actions in the main toolbar.
        tool_bar_actions = main_window.toolbar.actions()

        # For each such action...
        for first_separator in tool_bar_actions:
            # If this action is the first separator in this toolbar, this action
            # is the desired separator.
            if first_separator.isSeparator():
                # Insert a new separator before this separator in this toolbar,
                # such that:
                #
                # * The former separator will precede the undo and redo actions.
                # * The latter separator will succeed the undo and redo actions.
                main_window.toolbar.insertSeparator(first_separator)

                # Cease searching.
                break
        # Else, this action contains no separators. In this case, default to a
        # placeholder integer instructing subsequent insertions to append to
        # this toolbar. While this edge case should never happen, never is still
        # the enemy of reason.
        else:
            first_separator = 0

        # Insert undo and redo actions before this separator in this toolbar.
        main_window.toolbar.insertAction(first_separator, self._undo_action)
        main_window.toolbar.insertAction(first_separator, self._redo_action)

    # ..................{ PUSHERS                            }..................
    def push(self, undo_command: QUndoCommand) -> None:

        # Log this push *BEFORE* performing this push, for debuggability.
        logs.log_debug(
            'Pushing undo command "%s" onto stack...',
            undo_command.actionText())

        # logs.log_debug(
        #     'Action state *BEFORE*: undo (%r), redo (%r)',
        #     self._undo_action.isEnabled(),
        #     self._redo_action.isEnabled())

        # Push this undo command onto this stack.
        super().push(undo_command)

        # logs.log_debug(
        #     'Action state *AFTER*: undo (%r), redo (%r)',
        #     self._undo_action.isEnabled(),
        #     self._redo_action.isEnabled())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all :mod:`PySide2`-based controller subclasses.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QObject
from PySide2.QtWidgets import QMainWindow
from betse.util.type.types import type_check
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ SUPERCLASSES                       }....................
class QBetseeControllerABC(QBetseeObjectMixin, QObject):
    '''
    Abstract base class of all :mod:`PySide2`-based controller subclasses in the
    standard model-view-controller (MVC) understanding of that term.

    Each instance of this class is a controller encapsulating all abstract state
    (including connective logic like signals and slots) required to sanely
    display a separate physical view (i.e., widget). To minimally integrate with
    Qt concurrency and signalling, this controller is a minimal :class:`QObject`
    instance rather than a full-fledged :class:`QWidget` instance.
    '''

    # ..................{ INITIALIZERS                       }..................
    # To avoid circular import dependencies, this parameter is validated to be
    # an instance of the "QMainWindow" superclass rather than the expected
    # "QBetseeMainWindow" subclass of the "betsee.gui.window.guimainwindow"
    # submodule. Why? Because the latter imports the cached "betsee_ui.py"
    # module which imports the current submodule. Since this application only
    # contains one main window, this current validation suffices.
    @type_check
    def init(self, main_window: QMainWindow) -> None:
        '''
        Initialize this controller against the passed parent main window.

        To avoid circular references, this method is guaranteed to *not* retain
        a reference to this main window on returning. References to child
        widgets (e.g., simulation configuration stack widget) of this window may
        be retained, however.

        Parameters
        ----------
        parent: QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this widget.
        '''

        # Initialize our superclass.
        super().init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all editable enumerative simulation configuration
widget subclasses instantiated in pages of the top-level stack.
'''

#FIXME: Don't neglect to submit our "uiparser.py" fix to upstream! Nargle!

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, QObject  #, Signal, Slot
# from betse.util.io.log import logs
from betse.util.type.mapping import mappings
from betse.util.type.types import (
    type_check, ClassOrNoneTypes, EnumClassType, MappingType)
from betsee.guiexception import BetseePySideWidgetEnumException
from betsee.gui.simconf.stack.widget.abc.guisimconfwdgeditscalar import (
    QBetseeSimConfEditScalarWidgetMixin)

# ....................{ MIXINS                             }....................
class QBetseeSimConfEditEnumWidgetMixin(QBetseeSimConfEditScalarWidgetMixin):
    '''
    Abstract base class of all **editable enumerative simulation configuration
    widget** (i.e., widget interactively selecting between the mutually
    exclusive members of a simulation configuration enumeration stored in
    external YAML files) subclasses.

    Attributes
    ----------
    _enum_member_to_widget_value : MappingType
        Dictionary mapping from each member of the enumeration constraining this
        widget to the corresponding mutually exclusive value displayed by this
        widget.
    _widget_value_to_enum_member : MappingType
        Dictionary mapping from each mutually exclusive value displayed by this
        widget to the corresponding member of the enumeration constraining this
        widget.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._enum_member_to_widget_value = None
        self._widget_value_to_enum_member = None


    @type_check
    def init(
        self, enum_member_to_widget_value: MappingType, *args, **kwargs
    ) -> None:
        '''
        Finalize the initialization of this widget.

        Parameters
        ----------
        enum_member_to_widget_value : MappingType
            Dictionary mapping from each member of the enumeration encapsulated
            by the passed ``sim_conf_alias`` parameter to the corresponding
            mutually exclusive value displayed by this widget.

        All remaining parameters are passed as is to the superclass method.

        Raises
        ----------
        BetseMappingException
            If this dictionary is *not* safely invertible (i.e., if any value of
            this dictionary is non-uniquely assigned to two or more keys).
        BetseePySideRadioButtonException
            If the number of members in this enumeration differs from the number
            of members mapped by (i.e., of keys in) this dictionary.
        '''

        # Initialize our superclass with all remaining parameters.
        super().init(*args, **kwargs)

        # Classify both this dictionary and its inverse.
        self._enum_member_to_widget_value = enum_member_to_widget_value
        self._widget_value_to_enum_member = mappings.invert_map_unique(
            enum_member_to_widget_value)
        # logs.log_debug('enum_member_to_widget_value: {!r}'.format(
        #     self._enum_member_to_widget_value))
        # logs.log_debug('widget_value_to_enum_member: {!r}'.format(
        #     self._widget_value_to_enum_member))

        # Enumeration constraining this simulation configuration alias. Note
        # that this type need *NOT* validated here, as the
        # _sim_conf_alias_type_strict() property instructs our superclass to do
        # so already.
        enum_type = self._sim_conf_alias.data_desc.expr_alias_cls

        # If the number of members in this enumeration differs from the number
        # of members mapped by this dictionary, raise an exception.
        if len(enum_type) != len(enum_member_to_widget_value):
            raise BetseePySideWidgetEnumException(QCoreApplication.translate(
                'QBetseeSimConfEditEnumWidgetMixin',
                'Number of enumeration members {0} differs from '
                'number of mapped enumeration members {1}.'.format(
                    len(enum_type), len(enum_member_to_widget_value))))

    # ..................{ PROPERTIES                         }..................
    @property
    def _sim_conf_alias_type_strict(self) -> ClassOrNoneTypes:

        # Enumeration types are subclasses of the "EnumClassType" superclass and
        # instances of the "EnumType" type. Since the superclass
        # _die_if_sim_conf_alias_type_invalid() method validates types with
        # issubclass() rather than isinstance(), the former type is returned.
        return EnumClassType

    # ..................{ GETTERS                            }..................
    def _get_alias_from_widget_value(self) -> object:

        # Currently displayed value of this widget.
        widget_value = self.widget_value

        # If this value corresponds to no enumeration member, raise an
        # exception. While this should *NEVER* be the case, should should never
        # be a word.
        if widget_value not in self._widget_value_to_enum_member:
            # Human-readable description of this value.
            widget_value_label = widget_value

            # If this value is a Qt-specific object, reduce this description to
            # this object's name.
            if isinstance(widget_value, QObject):
                widget_value_label = widget_value.objectName()

            # Raise this exception with this description.
            raise BetseePySideWidgetEnumException(QCoreApplication.translate(
                'QBetseeSimConfEditEnumWidgetMixin',
                'Widget value "{0}" unrecognized.'.format(
                    widget_value_label)))

        # Return the enumeration member corresponding to this button.
        return self._widget_value_to_enum_member[widget_value]


    def _get_widget_from_alias_value(self) -> object:

        # Current value of this simulation configuration alias.
        enum_member = self._sim_conf_alias.get()

        # If this is *NOT* a member of this enumeration, raise an exception.
        # While this should *NEVER* be the case, should should never be a word.
        if enum_member not in self._enum_member_to_widget_value:
            raise BetseePySideWidgetEnumException(QCoreApplication.translate(
                'QBetseeSimConfEditEnumWidgetMixin',
                'Enumeration member "{0}" unrecognized.'.format(
                    str(enum_member))))

        # Return the radio button corresponding to this member.
        return self._enum_member_to_widget_value[enum_member]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.

'''
Low-level :mod:`PySide2`-specific logging configuration.
'''

#FIXME: Install a new PySide2.QtCore.qInstallMessageHandler() redirecting all
#Qt-specific messages to the standard Python-specific logging API. When we do,
#post the resulting solution to *BOTH* of the following:
#
#    https://stackoverflow.com/questions/35894171/redirect-qdebug-output-to-file-with-pyqt5
#    https://forum.qt.io/topic/84022/code-for-qinstallmessagehandler/8

#FIXME: Consider colourizing redirected log records. Doing so will be slightly
#non-trivial, however, as we would then need to use a rich text edit widget
#rather than a plain text edit widget.

# ....................{ IMPORTS                            }....................
from betse.util.io.log import logconfig, logs
from betse.util.io.log.logenum import LogLevel
from betse.util.io.log.logfilter import LogFilterThirdPartyDebug
from betse.util.type.types import type_check
from betsee.util.io.log.guiloghandle import LogHandlerSignal
from betsee.util.widget.stock.guitextedit import QBetseePlainTextEdit

# ....................{ INITIALIZERS                       }....................
@type_check
def log_to_text_edit(text_edit: QBetseePlainTextEdit) -> None:
    '''
    Append all unfiltered log records to the passed text widget in an
    autoscrolling, non-blocking, thread-safe manner.

    This function integrates the default logging configuration for the active
    Python process with the current :mod:`PySide2` application. Specifically,
    this function reconfigures logging to additionally forward all relevant log
    records logged to the root handler onto the relevant slot connected to the
    relevant signal of the passed text widget, where "relevant log records"
    means:

    * If verbosity is enabled (e.g., via the ``--verbose`` command-line option),
      all log records.
    * Else, all log records with level ``LogLevel.INFO`` and higher.

    Parameters
    ----------
    text_edit : QBetseePlainTextEdit
        Text widget to append relevant log records to.
    '''

    # Global logging configuration.
    log_config = logconfig.get()

    # Root logger handler redirecting to all slots connected to a signal.
    logger_root_handler_signal = LogHandlerSignal(
        signal=text_edit.append_text_signal)

    # If verbosity is enabled, redirect all log records; else, only redirect log
    # records with level "INFO" and higher.
    logger_root_handler_signal.setLevel(
        LogLevel.ALL if log_config.is_verbose else LogLevel.INFO)

    # Avoid redirecting third-party debug messages.
    logger_root_handler_signal.addFilter(LogFilterThirdPartyDebug())

    # Register this handler with the root logger.
    log_config.logger_root.addHandler(logger_root_handler_signal)

    # Log a single-line sentence synopsizing the state of this application
    # *AFTER* finalizing logging reconfiguration.
    logs.log_info('Initializing GUI...')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulator proactor** (i.e., :mod:`PySide2`-based object
controlling but *not* displaying the execution of simulation phases)
functionality.
'''

#FIXME: Improve the underlying exporting simulation subcommands in BETSE (e.g.,
#"betse plot seed") to perform routine progress callbacks.

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QCoreApplication, QObject, Slot  #, Signal
from PySide2.QtWidgets import QProgressBar, QLabel
from betse.exceptions import BetseSimUnstableException
from betse.util.io.log import logs
from betse.util.py import pythread
from betse.util.type import enums
from betse.util.type.obj import objects
from betse.util.type.types import type_check
from betsee.guiexception import (
    BetseeSimmerException, BetseeSimmerBetseException)
from betsee.gui.window.guimainwindow import QBetseeMainWindow
from betsee.gui.simtab.run.guisimrunenum import SimmerState
from betsee.gui.simtab.run.guisimrunstate import (
    SIMMER_STATES_IDLE,
    SIMMER_STATES_INTO_FIXED,
    SIMMER_STATES_FROM_FLUID,
    SIMMER_STATES_RUNNING,
    SIMMER_STATES_WORKING,
    SIMMER_STATES_UNWORKABLE,
)
from betsee.gui.simtab.run.guisimrunabc import QBetseeSimmerStatefulABC
from betsee.gui.simtab.run.phase.guisimrunphase import QBetseeSimmerPhase
from betsee.gui.simtab.run.phase.guisimrunphaser import QBetseeSimmerPhaser
from betsee.gui.simtab.run.work.guisimrunwork import QBetseeSimmerPhaseWorker
from betsee.util.thread import guithread
from betsee.util.thread.pool import guipoolthread
from collections import deque

# ....................{ CLASSES                           }....................
class QBetseeSimmerProactor(QBetseeSimmerStatefulABC):
    '''
    High-level **simulator proactor** (i.e., :mod:`PySide2`-based object
    controlling but *not* displaying the execution of simulation phases).

    This proactor maintains all state needed to manage these phases, including:

    * A queue of all simulation subcommands to be interactively run.
    * Whether or not a simulation subcommand is currently being run.
    * The state of the currently run simulation subcommand (if any), including:

      * Visualization (typically, Vmem animation) of the most recent step
        completed for this subcommand.
      * Textual status describing this step in human-readable language.
      * Numeric progress as a relative function of the total number of steps
        required by this subcommand.

    Design
    ----------
    This proactor implements the well-known `proactor design pattern`_, whereby
    each simulation subcommand is asynchronously run by a simulator worker
    assigned to a pooled thread dedicated to that worker. This pattern is
    effectively an asynchronous variant of the `reactor design pattern`_.

    For maintainability, this proactor subsumes *all* roles defined by this
    pattern except the core role of the **asynchronous operation** (e.g., work
    performed by each worker), including the following roles:

    * **Proactive initiator** (i.e., the highest-level parent object
      responsible for governing all asynchronous activity).
    * **Asynchronous operation processor** (i.e., the second-highest-level
      parent object responsible for starting the asynchronous activity and
      forwarding responsibility for handling its completion to the completion
      dispatcher).
    * **Completion dispatcher** (i.e., the second-lowest-level child object
      responsible for receiving control from the asynchronous operation
      processor on the completion of the asynchronous activity and forwarding
      responsibility for handling its completion to the completion handler).
    * **Completion handler** (i.e., the lowest-level child object responsible
      for handling the actual completion of the asynchronous activity).

    .. _proactor design pattern:
       https://en.wikipedia.org/wiki/Proactor_pattern
    .. _reactor design pattern:
       https://en.wikipedia.org/wiki/Reactor_pattern

    Caveats
    ----------
    For simplicity, this proactor internally assumes the active Python
    interpreter to prohibit Python-based multithreading via a Global
    Interpreter Lock (GIL). Specifically, all worker-centric attributes (e.g.,
    :attr:`_worker`, :attr:`_workers_queued`) are assumed to be implicitly
    synchronized despite access to these attributes *not* being explicitly
    locked behind a Qt-based mutual exclusion primitive.

    GIL-less Python interpreters violate this simplistic assumption. For
    example, the :meth:`stop_workers` and :meth:`_handle_worker_completion`
    slots suffer obvious (albeit unlikely) race conditions under GIL-less
    interpreters due to competitively deleting the same underlying worker in a
    desynchronized and hence non-deterministic manner.

    Yes, this is assuredly a bad idea. Yes, this is us nonchalantly shrugging.

    Attributes (Public)
    ----------
    phaser : QBetseeSimmerPhaser
        Container containing all simulator phase controllers.

    Attributes (Private)
    ----------
    _p : Parameters
        Simulation configuration singleton.

    Attributes (Private: Thread)
    ----------
    _thread : QBetseeWorkerThread
        Thread controller owning all simulator workers (i.e.,
        :class:`QBetseeSimmerWorkerABC` instances responsible for running
        queued simulation subcommands in a multithreaded manner).
    _workers_queued : {QueueType, NoneType}
        **Simulator worker queue** (i.e., double-ended queue of each simulator
        worker to be subsequently run in a multithreaded manner by the parent
        proactor to run a simulation subcommand whose corresponding checkbox
        was checked at the time this queue was instantiated) if this simulator
        has started one or more such workers *or* ``None`` otherwise (i.e., if
        no such workers have been started).

    Attributes (Private: Widgets)
    ----------
    _progress_bar : QProgressBar
        Alias of the :attr:`QBetseeMainWindow.sim_run_player_progress` widget.
    _progress_status : QLabel
        Alias of the :attr:`QBetseeMainWindow.sim_run_player_status` label.
    _progress_substatus : QLabel
        Alias of the :attr:`QBetseeMainWindow.sim_run_player_substatus` label.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this simulator.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        #FIXME: Yeah... fix this, please. PyPy will probably make the GIL
        #optionally go away at some point. (At least, it certainly should.)

        # Raise an exception unless the active Python interpreter has a GIL, as
        # foolishly required by methods defined by this class.
        pythread.die_unless_gil()

        # Nullify all remaining instance variables for safety.
        self._p = None
        self._progress_bar = None
        self._progress_status = None
        self._workers_queued = None

        # Container of allsSimulator phase controllers.
        self.phaser = QBetseeSimmerPhaser(self)


    @type_check
    def init(
        self,
        main_window: QBetseeMainWindow,
        progress_bar: QProgressBar,
        progress_status: QLabel,
        progress_substatus: QLabel,
    ) -> None:
        '''
        Finalize this simulator's initialization, owned by the passed main
        window widget.

        This method connects all relevant signals and slots of *all* widgets
        (including the main window, top-level widgets of that window, and leaf
        widgets distributed throughout this application) whose internal state
        pertains to the high-level state of this simulator.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        progress_bar : QProgressBar
            Alias of the :attr:`QBetseeMainWindow.sim_run_player_progress`
            widget.
        progress_status : QLabel
            Alias of the :attr:`QBetseeMainWindow.sim_run_player_status` label.
        progress_substatus : QLabel
            Alias of the :attr:`QBetseeMainWindow.sim_run_player_substatus`
            label.
        '''

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Log this finalization.
        logs.log_debug('Sanitizing simulator proactor state...')

        # Classify variables of this main window required by this simulator.
        self._p = main_window.sim_conf.p

        # Classify all remaining passed parameters.
        self._progress_bar       = progress_bar
        self._progress_status    = progress_status
        self._progress_substatus = progress_substatus

        # Initialize the container of all simulator phase controllers.
        self.phaser.init(
            main_window=main_window,
            set_state_from_phase=self._set_state_from_phase,
        )

    # ..................{ FINALIZERS                        }..................
    def halt_workers(self) -> None:
        '''
        Coercively (i.e., non-gracefully) halt the current simulator worker if
        any *and* dequeue all subsequently queued workers in a thread-safe
        manner, reverting the simulator to the idle state... **by any means
        necessary.**

        This high-level method subsumes the lower-level :meth:`stop_workers`
        slot by (in order):

        #. If no worker is currently working, silently reducing to a noop.
        #. Attempting to gracefully halt the currently working worker, dequeue
           all subsequently queued workers if any, and unblock this worker's
           parent thread if currently blocked.
        #. If this worker fails to gracefully halt within a reasonable window
           of time (e.g., 30 seconds), coerce this worker to immediately halt.

        Design
        ----------
        This method *must* called at application shutdown (e.g., by the parent
        main window). If this is not done *and* a previously running simulator
        worker is currently paused and hence indefinitely blocking its parent
        thread, this application will itself indefinitely block rather than
        actually shutdown. Which, of course, would be catastrophic.

        Caveats
        ----------
        This method may induce data loss or corruption in simulation output.
        In theory, this should only occur in edge cases in which the current
        simulator worker fails to gracefully stop within a sensible window of
        time. In practice, this implies that this method should *only* be
        called when otherwise unavoidable (e.g., at application shutdown).
        '''

        # Maximum number of milliseconds to block the current thread waiting
        # for the currently working worker (if any) to gracefully halt.
        WAIT_MAX_MILLISECONDS = 30000  # 30 seconds

        # Log this shutdown.
        logs.log_debug('Finalizing simulator workers...')

        # If no worker is currently working, silently reduce to a noop.
        if not self.is_worker:
            return
        # Else, some worker is currently working.

        # Currently working simulator worker. For safety, this property is
        # localized *BEFORE* this worker's stop() pseudo-slot (which
        # internally dequeues this worker and hence implicitly modifies the
        # worker returned by the "_worker" property) is called.
        worker = self.worker

        # Attempt to gracefully halt this worker, dequeue all subsequently
        # queued workers if any, and unblock this worker's parent thread if
        # currently blocked.
        self.stop_workers()

        # If this worker fails to gracefully halt within a reasonable window
        # of time (e.g., 30 seconds), coerce this worker to immediately halt.
        guipoolthread.halt_workers(
            workers=(worker,), milliseconds=WAIT_MAX_MILLISECONDS)

    # ..................{ PROPERTIES ~ bool                 }..................
    @property
    def is_queued(self) -> bool:

        # Return true only if one or more phases are queued.
        return any(phase.is_queued for phase in self.phaser.PHASES)


    @property
    def is_worker(self) -> bool:
        '''
        ``True`` only if one or more simulator workers currently exist.

        Equivalently, this property returns ``True`` only if the simulator
        worker queue is currently non-empty.

        Design
        ----------
        For safety, this property should be tested *before* attempting to
        access the :meth:`worker` property, which raises an exception when no
        simulator workers currently exist.
        '''

        # Return true only if a non-empty queue of all phases to be run
        # currently exists. The lifecycle of this queue is managed in a just in
        # time (JIT) manner corresponding exactly to whether or not the
        # simulator and hence a simulator worker is currently working.
        # Specifically:
        #
        # * This queue defaults to "None".
        # * The toggle_work() slot instantiates this queue on initially
        #   starting a new queue of simulator workers.
        # * The stop_workers() slot reverts this queue back to "None".
        #
        # For efficiency, return this queue reduced to a boolean -- equivalent
        # to this less efficient (but more readable) pair of tests:
        #
        #    return self._workers_cls is not None and len(self._workers_cls)
        return bool(self._workers_queued)

    # ..................{ PROPERTIES ~ bool : state         }..................
    @property
    def is_idle(self) -> bool:
        '''
        ``True`` only if this proactor is **idle** (i.e., *no* proactor
        worker is either working, stopping, or recently finished, implying this
        proactor to be in either the queued or unqueued states).

        Equivalently, this property returns ``True`` only if this proactor is
        currently performing no work such that the user has either queued or
        unqueued at least one phase since the last work performed by this
        proactor (if any) completed. In this case, this proactor has no
        meaningful data or metadata concerning current or recently completed
        work to display to the user and hence is genuinely "idle."
        '''

        return self.state in SIMMER_STATES_IDLE


    @property
    def is_workable(self) -> bool:
        '''
        ``True`` only if this proactor is **workable** (i.e., currently capable
        of performing work).

        This property returns ``True`` only if this proactor is guaranteed to
        be safely startable, resumable, or pausable and hence "workable."
        Equivalently, this property returns ``True`` only if the current state
        of this proactor is neither of the following states:

        * Unqueued. In this state, this proactor is required to wait until the
          user queues at least one simulation subcommand.
        * Stopping. In this state, this proactor is required to wait until the
          currently stopping worker does so gracefully.

        In either case, this proactor is incapable of performing work until its
        state changes to any other state -- all of which support work.
        '''

        return self.state not in SIMMER_STATES_UNWORKABLE


    @property
    def is_working(self) -> bool:
        '''
        ``True`` only if the simulator is **working** (i.e., some proactor
        worker is either running or paused from running some queued proactor
        phase and hence is *not* finished).

        If this fine-grained property is ``True``, note that it is necessarily
        the case that the coarse-grained :attr:`is_workable` property is also
        ``True`` but that the reverse is *not* necessarily the case.
        Equivalently, this proactor is *always* workable when it is working but
        *not* necessarily working when it is workable (e.g., due to currently
        being queued but unstarted).
        '''

        return self.state in SIMMER_STATES_WORKING


    @property
    def is_running(self) -> bool:
        '''
        ``True`` only if the simulator is **running** (i.e., some simulator
        worker is currently modelling or exporting some queued simulator phase
        and hence is neither paused nor finished).

        If this fine-grained property is ``True``, note that it is necessarily
        the case that the coarse-grained :attr:`is_working` and
        :attr:`is_workable` properties are also ``True`` but that the reverse
        is *not* necessarily the case.  Equivalently, the simulator is *always*
        working when it is running but *not* necessarily running when it is
        working (e.g., due to currently being paused).
        '''

        return self.state in SIMMER_STATES_RUNNING

    # ..................{ PROPERTIES ~ worker               }..................
    @property
    def worker(self) -> QBetseeSimmerPhaseWorker:
        '''
        **Currently working worker** (i.e., :class:`QRunnable` instance
        currently modelling or exporting a previously queued simulation phase
        in another thread) if any *or* raise an exception otherwise (i.e., if
        no workers are currently working).

        Raises
        ----------
        BetseeSimmerException
            If no worker is currently working.
        '''

        # If no worker is working, raise an exception.
        self._die_unless_working()

        # Return the head worker of the worker queue.
        return self._workers_queued[0]

    # ..................{ PROPERTIES ~ private : bool       }..................
    @property
    def _is_paused(self) -> bool:
        '''
        ``True`` only if the simulator is **paused** (i.e., some simulator
        worker is currently paused while previously modelling or exporting some
        queued simulator phase and hence is neither running nor finished).
        '''

        return self.state is SimmerState.PAUSED

    # ..................{ PROPERTIES ~ private : phase      }..................
    @property
    def _worker_phase_state(self) -> SimmerState:
        '''
        State of the queued simulator phase that is currently **running**
        (i.e., either being modelled or exported by this simulator) if any *or*
        raise an exception otherwise.

        Caveats
        ----------
        For safety, this property should *only* be accessed when this queue is
        guaranteed to be non-empty (i.e., when the :meth:`is_worker` property
        is ``True``).

        Design
        ----------
        This property getter trivially reduces to a direct access of the
        :attr:`_worker.phase.state` instance variable, but exists to define the
        corresponding non-trivial property setter.

        Raises
        ----------
        BetseeSimmerException
            If no simulator phase is currently running (i.e.,
            :attr:`_workers_cls` is either ``None`` or empty).
        '''

        return self.worker.phase.state


    @_worker_phase_state.setter
    @type_check
    def _worker_phase_state(self, state: SimmerState) -> None:
        '''
        Set the state of the queued simulator phase that is currently
        **running** (i.e., either being modelled or exported by this simulator)
        if any to the passed state *or* raise an exception otherwise.

        This property setter may additionally set the state of this simulator
        to the passed state (depending on the current state of this simulator
        and the passed state). Doing so avoids subtle desynchronization issues
        between the state of this phase and this simulator. In particular, if a
        queued simulator phase is currently running, this simulator's state is
        guaranteed to be *always* the state of that phase.

        Caveats
        ----------
        For safety, this property should *only* be accessed when this queue is
        guaranteed to be non-empty (i.e., when the :meth:`is_worker` property
        is ``True``).

        Parameters
        ----------
        state : SimmerState
            New state to set this phase to.

        Raises
        ----------
        BetseeSimmerException
            If no simulator phase is currently running (i.e.,
            :attr:`_workers_cls` is either ``None`` or empty).
        '''

        # Currently working phase, localized purely for negligible efficiency.
        worker_phase = self.worker.phase

        # Log this setting.
        logs.log_debug(
            'Updating simulator phase "%s" state from "%s" to "%s"...',
            worker_phase.name,
            enums.get_member_name_lowercase(worker_phase.state),
            enums.get_member_name_lowercase(state))

        # Set the state of the currently running phase to this state.
        worker_phase.state = state

        # Possibly set the state of this proactor to the same state.
        self._set_state_from_phase(worker_phase)


    @Slot(QObject)
    def _set_state_from_phase(self, phase: QBetseeSimmerPhase) -> None:
        '''
        Context-sensitively set the current state of this proactor to the
        current state of the passed proactor phase.

        Parameters
        ----------
        phase : QBetseeSimmerPhase
            Proactor phase whose current state has been previously set.
        '''

        # New state to change this proactor to if any *OR* "None" otherwise.
        state_new = None

        # If the current state of either:
        #
        # * This phase is fixed (i.e., high-priority) and hence superceding the
        #   current state of this proactor...
        # * This proactor is fluid (i.e., low-priority) and hence superceded
        #   by the current state of this phase...
        if (
            phase.state in SIMMER_STATES_INTO_FIXED or
             self.state in SIMMER_STATES_FROM_FLUID
        ):
            # If this phase's new state is unqueued...
            if phase.state is SimmerState.UNQUEUED:
                # If one or more *OTHER* phases are still queued, ignore this
                # phase's change. Unqueueing is minimally low priority.
                if self.is_queued:
                    pass
                # Else, all phases are unqueued. Set the current state of this
                # simulator to also be unqueued.
                else:
                    state_new = SimmerState.UNQUEUED
            # Else, this phase's new state unconditionally takes precedence.
            # Set this proactor's current state to this phase's new state.
            else:
                state_new = phase.state
        # Else, silently preserve this proactor's current state as is.

        # If preserving the state of this proactor...
        if state_new is None:
            # Log this preservation.
            logs.log_debug(
                'Preserving simulator state as "%s"...',
                enums.get_member_name_lowercase(self.state))

            # Preserve the state of this proactor below. As "None" is *NOT* a
            # valid state, reuse the existing state as is.
            state_new = self.state
        # Else, the state of this simulator is being changed. In this case...
        else:
            # Log this change.
            logs.log_debug(
                'Updating simulator proactor state from "%s" to "%s"...',
                enums.get_member_name_lowercase(self.state),
                enums.get_member_name_lowercase(state_new))

        # Unconditionally change the state of this proactor to this new state,
        # regardless of whether this state is already this new state. Why?
        # Because the superclass state() setter internally signals all slots
        # connected to the "state_changed" for this proactor -- including
        # the QBetseeSimmer._update_widgets() slot.
        #
        # The question then reduces to: "Do widgets require updating even when
        # the current state of this proactor remains unchanged?" The answer, of
        # course, is: "Absolutely." Proactor state is a lossy coarse-grained
        # abstraction that fails to capture all possible simulator state. In
        # particular, proactor state remains unchanged on dequeueing exactly
        # one but *NOT* all queued simulator phases -- a common case clearly
        # requiring widgets (e.g., progress substatus) to be updated.
        self.state = state_new

    # ..................{ EXCEPTIONS                        }..................
    def _die_unless_workable(self) -> None:
        '''
        Raise an exception unless the simulator is **workable** (i.e.,
        currently capable of performing work).

        Raises
        ----------
        BetseeSimmerException
            If the simulator is unworkable.

        See Also
        ----------
        :meth:`is_workable`
            Further details.
        '''

        if not self.is_workable:
            raise BetseeSimmerException(QCoreApplication.translate(
                'QBetseeSimmerProactor',
                'Simulator not workable '
                '(i.e., not currently working and no phases queued).'))

    # ..................{ EXCEPTIONS ~ state                }..................
    def _die_unless_running(self) -> None:
        '''
        Raise an exception unless the simulator is **running** (i.e., some
        simulator worker is currently modelling or exporting some queued
        simulation phase and hence is neither paused nor finished).

        Raises
        ----------
        BetseeSimmerException
            If the simulator is *not* running, in which case the simulator is
            either paused or finished.

        See Also
        ----------
        :meth:`is_running`
            Further details.
        '''

        if not self.is_running:
            raise BetseeSimmerException(QCoreApplication.translate(
                'QBetseeSimmerProactor',
                'Simulator not running '
                '(i.e., either paused, finished, or not started).'))


    def _die_unless_paused(self) -> None:
        '''
        Raise an exception unless the simulator is **paused** (i.e., some
        simulator worker is currently paused while previously modelling or
        exporting some queued simulation phase and hence is neither running nor
        finished).

        Raises
        ----------
        BetseeSimmerException
            If the simulator is *not* paused, in which case the simulator is
            either running or finished.

        See Also
        ----------
        :meth:`_is_paused`
            Further details.
        '''

        if not self._is_paused:
            raise BetseeSimmerException(QCoreApplication.translate(
                'QBetseeSimmerProactor',
                'Simulator not paused '
                '(i.e., either running, finished, or not started).'))

    # ..................{ EXCEPTIONS ~ working              }..................
    def _die_if_working(self) -> None:
        '''
        Raise an exception if some simulator worker is currently working.

        Raises
        ----------
        BetseeSimmerException
            If some simulator worker is currently working.

        See Also
        ----------
        :meth:`is_worker`
            Further details.
        '''

        if self.is_worker:
            raise BetseeSimmerException(QCoreApplication.translate(
                'QBetseeSimmerProactor', 'Simulation currently working.'))


    def _die_unless_working(self) -> None:
        '''
        Raise an exception unless some simulator worker is currently working.

        Raises
        ----------
        BetseeSimmerException
            If no simulator worker is currently working.

        See Also
        ----------
        :meth:`is_worker`
            Further details.
        '''

        if not self.is_worker:
            raise BetseeSimmerException(QCoreApplication.translate(
                'QBetseeSimmerProactor', 'No simulation currently working.'))

    # ..................{ SLOTS ~ action : work             }..................
    # Slots connected to signals emitted by "QAction" objects specific to the
    # currently running simulator worker if any.

    #FIXME: Rewrite documentation to note the considerably changed behaviour.
    @Slot(bool)
    def toggle_work(self, is_playing: bool) -> None:
        '''
        Slot signalled on either the user interactively *or* the codebase
        programmatically pushing the :class:`QPushButton` widget corresponding
        to the :attr:`_action_toggle_work` variable.

        This slot runs the currently queued phase by either:

        * If this phase is currently paused, resuming this phase.
        * Else, starting this phase.

        Parameters
        ----------
        is_playing : bool
            ``True`` only if the corresponding :class:`QPushButton` widget is
            toggled, implying the user to have requested that the simulator be
            either started or resumed, contextually depending on the current
            state of the simulator; conversely, ``False`` implies a request
            that the simulator be paused.
        '''

        # Log this slot.
        guithread.log_debug_thread_main(
            'Toggling simulator work by user request...')

        # If the simulator is unworkable, raise an exception.
        self._die_unless_workable()

        # If the user requested the simulator be started or unpaused...
        if is_playing:
            # If the simulator is currently working, some simulator worker was
            # previously working before being paused by a prior call to this
            # slot. In this case, resume this worker.
            if self.is_worker:
                self._resume_worker()
            # Else, the simulator is *NOT* currently working. In this case,
            # enqueue all currently queued simulation phases as simulator
            # workers and start the first such worker.
            else:
                self._start_workers()
        # Else, the user requested the simulator be paused. Do so.
        else:
            self._pause_worker()


    def _start_workers(self) -> None:
        '''
        Enqueue one simulator worker for each simulation subcommand whose
        corresponding checkbox in a simulator phase is currently checked *and*
        iteratively start each such worker in a thread-safe manner.

        Raises
        ----------
        BetseeSimmerException
            If either:

            * No simulator phase is currently queued (i.e., no such checkboxes
              are currently checked).
            * One or more workers are already working.
        '''

        # Log this action.
        guithread.log_debug_thread_main(
            'Starting simulator work by user request...')

        # If no simulator phase is currently queued, raise an exception.
        self._die_unless_queued()

        # If one or more workers are already working, raise an exception.
        guipoolthread.die_if_working()

        # Initialize the queue of simulator phases to be run.
        self._enqueue_workers()

        # Initiate iteration by starting the first enqueued worker and
        # connecting the stop signal emitted by that worker to a slot
        # iteratively repeating this process.
        self._loop_worker()


    def _pause_worker(self) -> None:
        '''
        Pause the currently running simulator.

        This method temporarily pauses the current simulator worker in a
        thread-safe manner safely resumable at any time by calling the
        :meth:`_resume_worker` method.

        Raises
        ----------
        BetseeSimmerException
            If the simulator is *not* currently running.
        '''

        # Log this action.
        guithread.log_debug_thread_main(
            'Pausing simulator work by user request...')

        # If the simulator is *not* currently running, raise an exception.
        self._die_unless_running()

        # Set both this proactor and the currently working phase to the paused
        # state *BEFORE* successfully pausing this worker. See the
        # stop_workers() method for related commentary on this order of logic.
        self._worker_phase_state = SimmerState.PAUSED

        # Pause the currently running proactor worker.
        self.worker.pause()


    def _resume_worker(self) -> None:
        '''
        Resume the currently paused simulator.

        This method resumes the current simulator worker in a thread-safe
        manner after having been previously paused by a call to the
        :meth:`_pause_worker` method.

        Raises
        ----------
        BetseeSimmerException
            If the simulator is *not* currently paused.
        '''

        # Log this action.
        guithread.log_debug_thread_main(
            'Resuming simulator work by user request...')

        # If the simulator is *not* currently paused, raise an exception.
        self._die_unless_paused()

        # Revert both this proactor and the currently working phase to the
        # worker-specific running state *BEFORE* successfully resuming this
        # worker. See the stop_workers() method for related commentary on this
        # order of logic.
        self._worker_phase_state = self.worker.simmer_state

        # Resume the currently paused simulator worker.
        self.worker.resume()

    # ..................{ SLOTS ~ action : stop             }..................
    @Slot()
    def stop_workers(self) -> None:
        '''
        Slot signalled on the user interactively (but *not* the codebase
        programmatically) clicking the :class:`QPushButton` widget associated
        with the :attr:`_actionstop_workers` action.

        This method effectively reverts the simulator to the idle state in a
        thread-safe manner by (in order):

        #. Unpausing the current simulator worker if currently paused, thus
           unblocking this worker's parent thread if currently blocked.
        #. Gracefully halting this worker.
        #. Dequeueing all subsequently queued workers.

        Raises
        ----------
        BetseeSimmerException
            If no simulator worker is currently working.
        '''

        # Log this slot.
        guithread.log_debug_thread_main(
            'Stopping simulator work by user request...')

        # If no worker is currently working, raise an exception.
        self._die_unless_working()

        # Currently working simulator worker. For safety, this property is
        # localized *BEFORE* this worker's stop() pseudo-slot (which
        # implicitly signals the _handle_worker_completion() slot, which
        # internally dequeues this worker and hence implicitly modifies the
        # worker returned by the "_worker" property) is called.
        worker = self.worker

        # Set both this proactor and the currently working phase to the
        # stopping state *BEFORE* successfully stopping this worker.
        #
        # Doing so enforces mutual exclusivity from the end user perspective
        # with respect to proactor state. Specifically, setting this state
        # prior to all subsequent logic forces the UI to immediately reflect
        # this request to stop all work and hence prevents the user from
        # issuing any subsequent actions at odds with that request (e.g., to
        # pause or re-stop work) -- regardless of whether that request is
        # successfully fulfilled. If the order of these operations were
        # reversed (i.e., if this worker was stopped prior to setting this
        # state), a window of time would exist in which the UI failed to
        # reflect this request to stop and hence permitted the user to issue
        # subsequent actions at odds with that request. In short, this is sane.
        self._worker_phase_state = SimmerState.STOPPING

        # Dequeue *ALL* currently enqueued simulator workers scheduled to work
        # following the currently working worker. While the latter could
        # theoretically be dequeued as well, doing so would prevent the
        # _handle_worker_completion() slot transitively called by the call to
        # the stop() pseudo-slot below from setting the state of the phase
        # associated with the currently working worker. In short, this works.
        #
        # For negligible efficiency, this queue is recreated from scratch to
        # contain only this worker rather than iteratively popping all items
        # except the first from this queue (e.g., by calling the
        # queues.pop_left() function).
        self._workers_queued = deque((worker,))

        # Stop this worker *AFTER* dequeueing all currently enqueued workers
        # and setting this simulator state.
        #
        # While feasible, reversing this order of operations invites subtle
        # race conditions between this slot and the _handle_worker_completion()
        # slot signalled by this call, which calls the _loop_worker() method,
        # which calls the _dequeue_workers() method. Since the stop() method
        # called here signals the _handle_worker_completion() slot in a
        # multithreaded and hence non-deterministic manner, badness ensues.
        worker.stop()

    # ..................{ QUEUERS                           }..................
    def _enqueue_workers(self) -> None:
        '''
        Create the **simulator worker queue** (i.e., :attr:`_workers_queued`
        variable) as specified by the pair of checkboxes associated with each
        simulator phase.

        This method enqueues (i.e., pushes onto this queue) workers in
        simulation phase order, defined as the ordering of the members of the
        :class:`betse.science.phase.phaseenum.SimPhaseKind` enumeration.
        Callers may safely run the simulation phases performed by these workers
        merely by sequentially assigning each worker enqueued in this queue to
        a thread via the
        :func:`betsee.util.thread.pool.guipoolthread.start_worker` function.

        For example:

        #. The :class:`QBetseeSimmerSubcommandWorkerModelSeed` worker (if any)
           is guaranteed to be queued *before*...
        #. The :class:`QBetseeSimmerSubcommandWorkerModelInit` worker (if any)
           is guaranteed to be queued *before*...
        #. The :class:`QBetseeSimmerSubcommandWorkerModelSim` worker (if any).

        Raises
        ----------
        BetseeSimmerException
            If either:

            * No simulator phase is currently queued.
            * Some simulator phase is currently running (i.e.,
              :attr:`_workers_queued` is already defined to be non-``None``).
        '''

        # Log this action.
        guithread.log_debug_thread_main('Enqueueing simulator workers...')

        # If this controller is *NOT* currently queued, raise an exception.
        self._die_unless_queued()

        # If some simulator worker is currently working, raise an exception.
        self._die_if_working()

        # Queue of all simulator workers to be subsequently run.
        self._workers_queued = self.phaser.enqueue_phase_workers()


    def _dequeue_workers(self) -> None:
        '''
        Revert the :attr:`_workers_queued` to ``None``, effectively dequeueing
        (i.e., popping) all previously queued simulator workers.
        '''

        # Log this action.
        guithread.log_debug_thread_main('Dequeueing simulator workers...')

        # If no simulator worker is currently working, raise an exception.
        self._die_unless_working()

        # Clear this queue, implicitly scheduling all previously queued workers
        # for garbage collection *AND* disconnecting all external slots
        # previously connected to signals defined by these workers.
        self._workers_queued = None

    # ..................{ WORKERS ~ loop                    }..................
    def _loop_worker(self) -> None:
        '''
        Iteratively run the next enqueued simulator worker if any *or* cleanup
        after this iteration otherwise (i.e., if no workers remain to be run).

        This method perform the equivalent of the body of the abstract loop
        iteratively starting and running all enqueued simulator workers.
        Specifically, this method iteratively starts the next simulator worker
        (i.e., head item of the :attr:`_workers_queued`) enqueued by a prior call
        to the :meth:`_enqueue_workers` method if this queue is non-empty
        *or* garbage-collects this queue otherwise (i.e., if already empty).

        Design
        ----------
        Ideally, the body of this method would be the body of a simple loop
        over all enqueued simulator workers. Since the
        :meth:`_handle_worker_completion` slot calling this method is only
        iteratively signalled by Qt on the completion of each worker, however,
        "rolling" this method into a loop is effectively infeasible.

        Technically, refactoring this method into a continuation-based
        generator would probably suffice to "roll" this method into a loop.
        Doing so, however, would require the use of an asynchronous
        Python-based event loop *and* a heavyweight architectural redesign. In
        short, the current approach stands as the most reasonable.
        '''

        # If no workers remain to be run, gracefully halt this iteration by
        # silently reducing to a noop.
        if not self.is_worker:
            # Log this halt.
            guithread.log_debug_thread_main(
                'Ceasing simulator worker iteration...')

            # Reduce to a noop.
            return
        # Else, one or more workers remain to be run.

        # Next worker to be run, localized for negligible efficiency.
        worker = self.worker

        # Log this work attempt.
        guithread.log_debug_thread_main('Iterating simulator worker...')

        # Set the state of both this proactor and the currently working phase
        # *BEFORE* successfully starting this worker. See the stop_workers()
        # method for related commentary on this order of logic.
        self._worker_phase_state = worker.simmer_state

        # Finalize this worker's initialization.
        worker.init(
            conf_filename=self._p.conf_filename,
            progress_bar=self._progress_bar,
            progress_label=self._progress_substatus,
            handler_failed=self._handle_worker_exception,
            handler_finished=self._handle_worker_completion,
        )

        # Start this worker *AFTER* establishing all signal-slot connections.
        guipoolthread.start_worker(worker)

    # ..................{ WORKERS ~ slot                    }..................
    # Slots connected to signals emitted by "QRunnable" workers.

    @Slot(Exception)
    def _handle_worker_exception(self, exception: Exception) -> None:
        '''
        Slot signalled on the currently running simulator worker erroneously
        raising an unexpected exception.

        This slot trivially handles this exception by re-raising this
        exception.  Since the only means of explicitly re-raising an exception
        exposed by Python 3.x is to encapsulate that exception inside another
        exception, this slot unconditionally raises a
        :class:`BetseeSimmerBetseException` exception encapsulating the passed
        exception.

        Parameters
        ----------
        exception : Exception
            Exception raised by this worker.

        Raises
        ----------
        BetseeSimmerBetseException
            Unconditionally encapsulates the passed exception.
        '''

        # Log this slot.
        guithread.log_debug_thread_main(
            'Catching simulator worker exception "%s"...',
            objects.get_class_name(exception))

        # If this exception signifies an instability, raise a human-readable
        # translated exception synopsizing this fact. This error is
        # sufficiently common to warrant a special case improving the user
        # experience (UX).
        if isinstance(exception, BetseSimUnstableException):
            raise BetseeSimmerBetseException(
                synopsis=QCoreApplication.translate(
                    'QBetseeSimmerProactor',
                    'Simulation halted prematurely '
                    'due to computational instability.'),
            ) from exception
        # Else, fallback to raising a human-readable translated exception
        # synopsizing an unrecognized fatal error.
        #
        # Note that this case embeds an exegesis (i.e., detailed message) *NOT*
        # embedded in the prior case and hence requires unique handling.
        else:
            raise BetseeSimmerBetseException(
                synopsis=QCoreApplication.translate(
                    'QBetseeSimmerProactor',
                    'Simulation halted prematurely with unexpected error:'),
                exegesis=str(exception),
            ) from exception


    @Slot(bool)
    def _handle_worker_completion(self, is_success: bool) -> None:
        '''
        Handle the completion of the most recently working simulator worker.

        Specifically, this method:

        * Sets the state of the corresponding simulator phase to finished.
        * Pops this worker from the :attr:`_workers_queued`.
        * If this queue is non-empty, starts the next enqueued worker.

        Parameters
        ----------
        is_success : bool
            ``True`` only if this worker completed successfully.
        '''

        # If the most recently working worker is no longer working, silently
        # reduce to a noop. Ideally, a worker would *ALWAYS* be working when
        # this slot is signalled. In practice, edge cases resulting from the
        # non-determinism implicit in multithreaded logic can induce this. For
        # example, the stop_workers() slot dequeueing all workers *BEFORE*
        # calling the QBetseeSimmerPhaseWorker.stop() pseudo-slot signalling
        # this slot reliably induces this case.
        if not self.is_worker:
            # Log this edge case.
            guithread.log_debug_thread_main(
                'Ignoring simulator worker closure...')

            # Reduce to a noop.
            return
        # Else, one or more workers remain to be run.

        # Log this slot.
        guithread.log_debug_thread_main('Handling simulator worker closure...')

        # If this worker is still technically working, set the state of both
        # this simulator *AND* the previously working phase to finished. This
        # ensures that the simulator reliably returns to the finished state on
        # completing all possible work, regardless of whether that worker
        # halted gracefully or prematurely (e.g., due to the user prematurely
        # stopping all work).
        #
        # For safety, do so *BEFORE* this method dequeues this worker and hence
        # internally modifies the worker yielded by the "_worker" property.
        self._worker_phase_state = SimmerState.FINISHED

        # Schedule this worker for immediate deletion. On doing so, all signals
        # owned by this worker will be disconnected from connected slots.
        #
        # In theory, doing so *SHOULD* be unnecessary. The subsequent
        # nullification of the worker owning these signals should already
        # signal this deletion. As doing so has no harmful side effects,
        # however, do so regardless for additional safety.
        self.worker.delete_later()

        # Dequeue this worker (i.e., remove this worker's subclass from
        # the queue of worker subclasses to be instantiated and run).
        self._workers_queued.popleft()

        # Start the next worker in the current queue of workers to be run
        # if any or reduce to a noop otherwise.
        self._loop_worker()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
General-purpose :mod:`QTreeWidget` subclasses.
'''

# ....................{ IMPORTS                           }....................
# from PySide2.QtCore import Signal, Slot
from PySide2.QtWidgets import QHeaderView, QTreeWidget, QTreeWidgetItem
from betse.util.type.types import type_check, GeneratorType
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ REMOVERS                          }....................
@type_check
def remove_item(item: QTreeWidgetItem) -> None:
    '''
    Remove the passed tree item from its parent tree item and hence the parent
    tree transitively containing those items.

    Parameters
    ----------
    item : QTreeWidgetItem
        Tree item to be removed from its parent tree item and tree widget.

    See Also
    ----------
    https://stackoverflow.com/a/8961820/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Defer heavyweight imports. Sadly, this submodule is only available under
    # properly packaged unstable releases of "PySide2". *sigh*
    from PySide2 import shiboken2

    # Yes, this actually removes this item from its parent tree. Why? Because
    # the high-level shiboken2.delete() function wraps the low-level C++
    # "delete" operator. Naturally, the "QTreeWidgetItem" class overrides the
    # "delete" operator to remove itself from its parent tree -- which, in C++,
    # is the established means of doing so. Ergo, this is the established means
    # of doing so in Python as well.
    shiboken2.delete(item)

# ....................{ SUBCLASSES                        }....................
class QBetseeTreeWidget(QBetseeObjectMixin, QTreeWidget):
    '''
    :mod:`QTreeWidget`-based widget marginally improving upon the stock
    :mod:`QTreeWidget` functionality.

    This application-specific widget augments the :class:`QTreeWidget` class
    with additional support for:

    * **Horizontal scrollbars,** automatically displaying horizontal scrollbars
      for all columns whose content exceeds that column's width. For
      inexplicable reasons, this functionality has been seemingly intentionally
      omitted from the stock :class:`QTreeWidget`.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Header view for this tree.
        header_view = self.header()

        # To display a horizontal scrollbar instead of an ellipse when resizing
        # a column smaller than its content, resize that column's section to
        # its optimal size. For further details, see the following FAQ entry:
        #     https://wiki.qt.io/Technical_FAQ#How_can_I_ensure_that_a_horizontal_scrollbar_and_not_an_ellipse_shows_up_when_resizing_a_column_smaller_than_its_content_in_a_QTreeView_.3F
        header_view.setSectionResizeMode(QHeaderView.ResizeToContents)

        #FIXME: Consider excising this. The following call once behaved as
        #expected under obsolete PySide2 releases but now appears to
        #prematurely truncate this column. Oddly, the default behaviour now
        #behaves as expected. Such is our forlorn life in the code trenches.

        # By default, all trees contain only one column. Under the safe
        # assumption this tree will continue to contain only one column,
        # prevent this column's content from automatically resizing to the
        # width of the viewport rather than this column's section (as requested
        # by the prior call). This unfortunate default overrides that request.
        # header_view.setStretchLastSection(False)

    # ..................{ SUPERCLASS ~ setters              }..................
    def setColumnCount(self, column_count: int) -> None:

        # Defer to the superclass implementation.
        super().setColumnCount(column_count)

        # If this tree now contains more than one column, permit the last such
        # column's content to automatically resize to the viewport width.
        if column_count != 1:
            self.header().setStretchLastSection(True)

    # ..................{ ITERATORS                         }..................
    def iter_items_top(self) -> GeneratorType:
        '''
        Generator iteratively yielding each **top-level tree item** (i.e.,
        :class:`QTreeWidgetItem` owned by this tree such that the parent item
        of this child item is the invisible root item returned by the
        :meth:`invisibleRootItem` method) of this tree widget.

        Yields
        ----------
        QTreeWidgetItem
            Current top-level tree item of this tree widget.

        See Also
        ----------
        https://stackoverflow.com/a/8961820/2809027
            StackOverflow answer strongly inspiring this implementation.
        '''

        # Root tree item of this tree widget.
        item_root = self.invisibleRootItem()

        # Number of top-level tree items (i.e., direct children of this parent
        # root tree item).
        items_top_count = item_root.childCount()

        # Return a generator comprehension yielding...
        return (
            # Top-level tree item with the current 0-based index...
            item_root.child(item_top_index)
            # For the 0-based index of each top-level tree item.
            for item_top_index in range(items_top_count)
        )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application-specific monkey patching of the :mod:`setuptools`-specific
:class:`ScriptWriter` class.

Such patching improves the usability of this class with respect to editable
installations packages (e.g., via this application's *nix-specific ``symlink``
subcommand or the general-purpose :mod:`setuptools ``develop`` subcommand). The
default :class:`ScriptWriter` implementation writes scripts attempting to
import the :mod:`setuptools`-installed package resources for these packages.
Since no such resources are installed for editable installations, these scripts
*always* fail and hence are suitable *only* for use in user-specific venvs.

Such patching corrects this deficiency, albeit at a minor cost of ignoring the
package resources provided by Python packages installed in the customary way.
While there exist alternatives, this appears to be the most robust means of
maintaining backward compatibility with older :mod:`setuptools` versions.
'''

# ....................{ IMPORTS                           }....................
from betsee_setup import beuputil
from distutils.errors import DistutilsClassError
from pkg_resources import Distribution
from setuptools.command import easy_install
from setuptools.command.easy_install import ScriptWriter, WindowsScriptWriter

# ....................{ CONSTANTS                         }....................
SCRIPT_TEMPLATE = '''
# This script is auto-generated by the "build_scripts" setuptools subcommand
# monkey-patched by the BETSEE-specific "betsee_setup.beupbuild" submodule.
#
# Welcome to the theatre of the absurd.

import importlib, sys

def is_module_root(module_name: str) -> bool:
    """
    ``True`` only if the Python module with the passed fully-qualified name is
    a **top-level module** (i.e., module whose name contains no ``.``
    delimiters) importable under the active Python interpreter.

    If this is *not* a top-level module, an exception is raised. If this
    top-level module is *not* importable via the standard
    :func:`importlib.find_loader` mechanism (e.g., the OS X-specific
    :mod:`PyObjCTools` package), this module may be imported by this function
    as an unwanted side effect.
    """
    assert isinstance(module_name, str), (
        '"{{}}" not a string.'.format(module_name))

    # If this is *NOT* a top-level module, raise an exception.
    if '.' in module_name:
        raise ImportError(
            'Module "{{}}" not a top-level module.'.format(module_name))

    # See betse.util.py.modules.is_module() for implementation details.
    try:
        return importlib.find_loader(module_name) is not None
    except ValueError:
        try:
            importlib.import_module(module_name)
            return True
        except ImportError:
            return False

# If this script is run directly, do so.
if __name__ == '__main__':
    # If the root parent package of this entry module is unimportable, raise a
    # human-readable exception. For inscrutable reasons, testing for whether
    # this entry module itself is importable is highly non-trivial under Python
    # 3.3 but *NOT* Python >= 3.4. While requiring Python >= 3.4 above would
    # obviate this, such version requirements are best asserted in the main
    # application codebase. Instead, we defer to the next best test.
    if not is_module_root('{entry_package_root}'):
        raise ImportError(
            'Package "{entry_package_root}" unimportable. '
            'Consider re-running either:\\n\\n'
            '\\tsudo python3 setup.py install\\n'
            '\\tsudo python3 setup.py develop\\n'
            '\\tsudo python3 setup.py symlink')

    # Import the entry module.
    import {entry_module} as entry_module

    # For debugging purposes, print the absolute path of this module.
    #print('{entry_module}: ' + entry_module.__file__)
    {entry_func_code}
# Else, this script is imported by another module rather than run directly. In
# this erroneous case, noop by printing a non-fatal warning and then returning.
# While this script should *NEVER* be imported, edge cases do happen.
else:
    print('WARNING: Entry point imported rather than run.', file=sys.stderr)
'''
'''
Script template to be formatted by the
:meth:`ScriptWriterSimple.get_script_args` method.
'''


SCRIPT_ENTRY_FUNC_SUBTEMPLATE = '''
    # If this module requires an entry function to be run, call this function.
    # For POSIX compliance, propagate the value returned by this function
    # (ideally a single-byte integer) back to the calling process as this
    # script's exit status.
    sys.exit(entry_module.{entry_func}())
'''
'''
Script subtemplate to be formatted by the
:meth:`ScriptWriterSimple.get_script_args` method for entry points requiring an
entry function to be called.

This excludes entry points for which merely importing the desired entry module
suffices to implicitly run that entry point -- typically, entry modules with
basename ``__main__``.
'''

# ....................{ COMMANDS                          }....................
def add_setup_commands(metadata: dict, setup_options: dict) -> None:
    '''
    Add commands building distribution entry points to the passed dictionary of
    :mod:`setuptools` options.
    '''

    # If neither of the class functions monkey-patched below exist, setuptools
    # is either broken or an unsupported newer version. In either case, an
    # exception is raised.
    if (not hasattr(ScriptWriter, 'get_args') and
        not hasattr(ScriptWriter, 'get_script_args')):
        raise DistutilsClassError(
            'Class "setuptools.command.easy_install.ScriptWriter" '
            'methods get_args() and get_script_args() not found. '
            'The current version of setuptools is either broken '
            '(unlikely) or unsupported (likely).'
        )

    # Monkey-patch the following class methods:
    #
    # * ScriptWriter.get_args(), defined by recent versions of setuptools.
    # * ScriptWriter.get_script_args(), defined by obsolete versions of
    #   setuptools.
    #
    # For convenience, our implementation of the latter is implemented in terms
    # of the former. Hence, the former is *ALWAYS* monkey-patched.
    ScriptWriter.get_args = _patched_get_args

    # If the ScriptWriter.get_script_args() class method exists, monkey-patch
    # both that *AND* the setuptools.command.easy_install.get_script_args()
    # alias referring to that method as well.
    if hasattr(ScriptWriter, 'get_script_args'):
        ScriptWriter.get_script_args = _patched_get_script_args
        easy_install.get_script_args = ScriptWriter.get_script_args

        # If the ScriptWriter.get_script_header() class method does *NOT*
        # exist, monkey-patch that method as well.
        if not hasattr(ScriptWriter, 'get_script_header'):
            ScriptWriter.get_script_header = _patched_get_script_header

# ....................{ PATCHES                           }....................
# Functions monkey-patching existing methods of the "ScriptWriter" class above
# and hence defined to have the same method signatures. The "cls" parameter
# implicitly passed to such methods by the @classmethod decorator is guaranteed
# to be the "ScriptWriter" class.

@classmethod
def _patched_get_args(
    cls: type,
    distribution: Distribution,
    script_shebang: str = None
):
    '''
    Yield ``write_script()`` argument tuples for the passed distribution's
    **entry points** (i.e., platform-specific executables running this
    distribution).

    This function monkey-patches the :meth:`ScriptWriter.get_args` class
    method.
    '''

    # Default this shebang line if unpassed.
    if script_shebang is None:
        script_shebang = cls.get_header()

    assert isinstance(cls, type), '"{}" not a class.'.format(cls)
    assert isinstance(script_shebang, str), (
        '"{}" not a string.'.format(script_shebang))
    #print('In BETSE ScriptWriter.get_args()!')

    # For each entry point of this distribution...
    for script_basename, script_type, entry_point in (
        beuputil.package_distribution_entry_points(distribution)):
        # If this entry point provides the name of the main function in this
        # entry module to be called, define script code calling this function.
        if len(entry_point.attrs):
            script_entry_func_code = SCRIPT_ENTRY_FUNC_SUBTEMPLATE.format(
                entry_func=entry_point.attrs[0])
        # Else, default this script code to the empty string.
        else:
            script_entry_func_code = ''

            # Print a non-fatal warning, as the resulting script may *NOT*
            # necessarily be runnable or freezable as expected.
            beuputil.output_warning(
                'Entry module "{}" entry function undefined.'.format(
                entry_point.module_name))

        # Script contents, formatted according to such template.
        script_code = SCRIPT_TEMPLATE.format(
            # Script code calling this entry module's main function.
            entry_func_code=script_entry_func_code,

            # Fully-qualified name of this entry module's root parent package.
            entry_package_root=entry_point.module_name.split('.')[0],

            # Fully-qualified name of this entry module.
            entry_module=entry_point.module_name,
        )

        # Yield a tuple containing this metadata to the caller. Note that the
        # _get_script_args() method called here is *NOT* the
        # _patched_get_script_args() method defined below. Confusing, but true.
        for script_tuple in cls._get_script_args(
            script_type, script_basename, script_shebang, script_code):
            yield script_tuple


@classmethod
def _patched_get_script_args(
    cls: type,
    distribution: Distribution,
    executable = None,
    is_windows_vanilla: bool = False
):
    '''
    Yield ``write_script()`` argument tuples for the passed distribution's
    **entry points** (i.e., platform-specific executables running this
    distribution).

    This function monkey-patches the deprecated
    :meth:`ScriptWriter.get_script_args` class method.
    '''
    assert isinstance(cls, type), '"{}" not a class.'.format(cls)
    # print('In BETSE ScriptWriter.get_script_args()!')

    # Platform-specific entry point writer.
    #
    # If the newer ScriptWriter.best() class function exists, obtain this
    # writer by calling this function.
    script_writer = None
    if hasattr(ScriptWriter, 'best'):
        script_writer = (
            WindowsScriptWriter if is_windows_vanilla else ScriptWriter).best()
    # Else, obtain this writer by calling the older ScriptWriter.get_writer()
    # class function.
    else:
        script_writer = cls.get_writer(is_windows_vanilla)

    # Shebang line prefixing the contents of this script.
    script_shebang = cls.get_script_header('', executable, is_windows_vanilla)

    # Defer to the newer get_args() method. Note that this is the
    # _patched_get_args() method defined above.
    return script_writer.get_args(distribution, script_shebang)


@classmethod
def _patched_get_script_header(cls: type, *args, **kwargs):
    '''
    Defer to the deprecated
    :func:`setuptools.command.easy_install.get_script_header` function under
    older :mod:setuptools versions.
    '''

    from setuptools.command.easy_install import get_script_header
    return get_script_header(*args, **kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based file functionality.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtWidgets import QFileDialog
from betse.util.path import dirs, pathnames, paths
from betse.util.type.iterable import sequences
from betse.util.type.text import strs
from betse.util.type.types import (
    type_check,
    CallableTypes,
    IntOrNoneTypes,
    MappingType,
    MappingOrNoneTypes,
    NoneType,
    StrOrNoneTypes,
)
from betsee.util.app import guiappwindow

# ....................{ CALLERS                            }....................
@type_check
def select_path(
    # Mandatory parameters.
    dialog_callable: CallableTypes,
    dialog_title: str,

    # Optional parameters.
    dialog_options: (QFileDialog.Option, NoneType) = None,
    init_pathname: StrOrNoneTypes = None,
    parent_dirname: StrOrNoneTypes = None,
    is_subpaths: bool = False,
    label_to_filetypes: MappingOrNoneTypes = None,
) -> StrOrNoneTypes:
    '''
    Absolute or relative pathname of a path interactively selected by the end
    user from a dialog displayed by calling the passed static getter of the
    :class:`QFileDialog` class (e.g., :func:`QFileDialog.getOpenFileName`)
    configured with the passed parameters if this dialog was not cancelled *or*
    ``None`` otherwise (i.e., if this dialog was cancelled).

    Constraints
    ----------
    Whether the pathname returned by this function is absolute or relative *and*
    whether the initial pathname accepted by this function is required to be
    absolute or relative depends on which optional parameters are passed.

    Specifically:

    * If the ``parent_dirname`` parameter is non-``None``:
      * The returned pathname is either:
        * If this path resides in this parent directory or a subdirectory
          thereof, relative to the absolute pathname of this parent directory.
        * Else, absolute.
      * If the ``init_pathname`` parameter is also non-``None``, this pathname
        may be either:
        * Relative, in which case this pathname is interpreted as relative to
          the absolute pathname of this parent directory.
        * Absolute, in which case this pathname is preserved as is.
    * Else (i.e., if the ``parent_dirname`` parameter is ``None``):
      * The returned pathname is absolute.
      * If the ``init_pathname`` parameter is non-``None``, this pathname *must*
        also be absolute. If this is *not* the case, an exception is raised.

    Parameters
    ----------
    dialog_callable : CallableTypes
        Static getter function of the :class:`QFileDialog` class to be called by
        this function (e.g., :func:`QFileDialog.getOpenFileName`).
    dialog_title : str
        Human-readable title of this dialog.
    dialog_options : (QFileDialog.Option, NoneType)
        *Bit field* (i.e., integer OR-ed together from mutually exclusive bit
        flags ala C-style enumeration types) of all :attr:`QFileDialog.Option`
        flags with which to configure this dialog. Defaults to ``None``, in
        which case this dialog defaults to the default configuration.
    init_pathname : StrOrNoneTypes
        Absolute or relative pathname of the path of arbitrary type (e.g., file,
        directory) to initially display in this dialog. Defaults to ``None``, in
        which case the pathname selected by the most recent call of this
        function is defaulted to.
    parent_dirname : StrOrNoneTypes
        Absolute pathname of the parent directory to select a path from.
        Defaults to ``None``, in which case no parental constraint is applied.
        See above for details.
    is_subpaths : bool
        ``True`` only if both the ``init_pathname`` parameter *and* the returned
        selected path are required to be children of the ``parent_dirname``
        parameter (i.e., residing either in this directory or a transitive
        subdirectory of this directory). Defaults to ``False``, in which case
        these paths may reside in any directory.
    label_to_filetypes : MappingOrNoneTypes
        Dictionary mapping from a human-readable label to be displayed for each
        iterable of related filetypes selectable by this dialog (e.g.,
        ``Images``) to that iterable (e.g., ``('jpg', 'png')``) if this dialog
        selects only files of specific filetypes *or* ``None`` otherwise. If
        this dialog only selects directories, this should be ``None``. Defaults
        to ``None``, in which case all paths of the requisite type (e.g., file,
        directory) regardless of filetype are selectable.

    Returns
    ----------
    StrOrNoneTypes
        Either:
        * If this dialog was confirmed, the absolute pathname of this path.
        * If this dialog was cancelled, ``None``.
    '''

    # Avoid circular import dependencies.
    from betsee.util.path import guipathsys

    # If no initial path was passed, default to a sane path. To avoid subtle
    # edge cases, do so *BEFORE* handling the passed parent directory if any.
    if init_pathname is None:
        init_pathname = guipathsys.get_path_dialog_init_pathname()

    # If no parent directory was passed...
    if parent_dirname is None:
        # If this initial pathname is relative, raise an exception. Since no
        # parent directory was passed, this pathname *CANNOT* be relativized.
        pathnames.die_if_relative(init_pathname)
    # Else, a parent directory was passed. In this case...
    else:
        # If this parent directory is relative, raise an exception.
        pathnames.die_if_relative(parent_dirname)

        # If this parent directory does *NOT* exist, raise an exception.
        dirs.die_unless_dir(parent_dirname)

        # If this initial path is relative, expand this into an absolute
        # pathname relative to this parent directory.
        if pathnames.is_relative(init_pathname):
            init_pathname = pathnames.join(parent_dirname, init_pathname)
        # Else, this initial path is absolute. If this path is required to live
        # in this parent directory, raise an exception if this is not the case.
        elif is_subpaths:
            pathnames.die_unless_parent(
                parent_dirname=parent_dirname, child_pathname=init_pathname)

    # If this initial path does *NOT* exist, silently reduce this path to the
    # last (i.e., most deeply nested) parent directory of this path that
    # exists. This ensures this dialog opens to an existing path.
    if not paths.is_path(init_pathname):
        init_pathname = dirs.get_parent_dir_last(init_pathname)

    # List of all arguments to be passed to this dialog creation function.
    dialog_args = [
        # Parent widget of this dialog.
        guiappwindow.get_main_window(),

        # Translated title of this dialog.
        dialog_title,

        # Initial path displayed by this dialog.
        init_pathname,
    ]

    # If a dictionary of acceptable filetypes was passed...
    if label_to_filetypes is not None:
        # String listing these filetypes converted from this dictionary.
        filetypes_filter = _make_filetypes_filter(label_to_filetypes)

        # Append this string to the tuple of all arguments to pass.
        dialog_args.append(filetypes_filter)

        # Append an explicit "None" value for the following optional
        # "selectedFilter" parameter, thus assigning all subsequently appended
        # positional parameters (e.g., "dialog_options") the proper index.
        dialog_args.append(None)

    # If a bit field of dialog flags was passed, append this field to the tuple
    # of all arguments to pass.
    if dialog_options is not None:
        dialog_args.append(dialog_options)
    # logs.log_debug('QFileDialog args: %r', dialog_args)

    # Absolute pathname of the path selected by the user if this dialog was not
    # canceled *OR* the empty string otherwise.
    pathname = dialog_callable(*dialog_args)

    # If this pathname is actually a 2-tuple "(pathname, filetypes_filter)",
    # this dialog was created by a file-specific callable (e.g.,
    # QFileDialog.getOpenFileName()) rather than directory-specific callable
    # (e.g., QFileDialog.getExistingDirectory()).
    #
    # Extract this pathname from the first item of this tuple. The second item
    # of this tuple is the "filetypes_filter" substring corresponding to the
    # filetype of this file (e.g., "YAML files (*.yml *.yaml);;" for a file
    # "sim_config.yaml"). Since this substring conveys no metadata not already
    # conveyed by this filetype, this substring is safely ignorable.
    if sequences.is_sequence(pathname) and len(pathname) == 2:
        pathname = pathname[0]

    # If this dialog was canceled, silently noop.
    if not pathname:
        return None
    # Else, this dialog was *NOT* canceled.

    # If a parent directory was passed...
    if parent_dirname is not None:
        # If this path lives in this parent directory, reduce this path to a
        # relative pathname relative to this parent directory.
        if pathnames.is_parent(
            parent_dirname=parent_dirname, child_pathname=pathname):
            pathname = pathnames.relativize(
                src_dirname=parent_dirname, trg_pathname=pathname)
        # ELse, this path does *NOT* live in this parent directory. If this path
        # is required to do so, raise an exception.
        elif is_subpaths:
            pathnames.die_unless_parent(
                parent_dirname=parent_dirname, child_pathname=init_pathname)

    # Return this pathname.
    return pathname

# ....................{ PRIVATE ~ makers                   }....................
@type_check
def _make_filetypes_filter(label_to_filetypes: MappingType) -> str:
    '''
    ``;;``-delimited string of all filetypes (e.g.,
    ``All files (*);; YAML files (*.yaml *.yml)``) specified by the passed
    dictionary, syntactically conforming to the Qt-specific format supported by
    the :class:`QFileDialog` class.

    Parameters
    ----------
    label_to_filetypes : MappingType
        Dictionary mapping from a human-readable label to be displayed for each
        iterable of related filetypes selectable by this dialog (e.g.,
        ``Images``) to that iterable (e.g., ``('jpg', 'png')``).

    Returns
    ----------
    str
        ``;;``-delimited string of all filetypes specified by this dictionary.
    '''

    # Qt-formatted string of the form accepted by "QFileDialog".
    filetypes_filter = ''

    # For each such filetype...
    for filetypes_label, filetypes_iterable in label_to_filetypes.items():
        # String listing these filetypes, delimited by spaces and...
        filetypes_listed = strs.join_on_space(
            # Prefixed by "*." for each filetype *NOT* equal to "*", as the
            # filetype "*.*" is more restrictive and hence less desirable
            # than the filetype "*".
            filetype if filetype == '*' else (
                '*' + pathnames.dot_filetype(filetype))
            for filetype in filetypes_iterable
        )

        # Qt-formatted string containing this label and these filetypes.
        filetypes_filter += '{} ({});;'.format(
            filetypes_label, filetypes_listed)

    # Return this Qt-formatted string.
    return filetypes_filter



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based stack widget page controllers specific to paths.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtWidgets import QMainWindow
from betse.science.parameters import Parameters
# from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfPathStackedWidgetPager(QBetseeControllerABC):
    '''
    :mod:`PySide2`-based stack widget page controller, connecting all editable
    widgets of the paths page with the corresponding low-level settings of the
    current simulation configuration.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def init(self, main_window: QMainWindow) -> None:

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Simulation configuration state object.
        sim_conf = main_window.sim_conf

        # Initialize all seed widgets on this page.
        main_window.sim_conf_path_seed_pick_file_line.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.seed_pickle_basename)

        # Initialize all initialization widgets on this page.
        main_window.sim_conf_path_init_pick_file_line.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.init_pickle_basename,
        )
        main_window.sim_conf_path_init_pick_dir_line.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.init_pickle_dirname_relative,
            push_btn=main_window.sim_conf_path_init_pick_dir_btn,
        )
        main_window.sim_conf_path_init_exp_dir_line.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.init_export_dirname_relative,
            push_btn=main_window.sim_conf_path_init_exp_dir_btn,
        )

        # Initialize all simulation widgets on this page.
        main_window.sim_conf_path_sim_pick_file_line.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.sim_pickle_basename,
        )
        main_window.sim_conf_path_sim_pick_dir_line.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.sim_pickle_dirname_relative,
            push_btn=main_window.sim_conf_path_sim_pick_dir_btn,
        )
        main_window.sim_conf_path_sim_exp_dir_line.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.sim_export_dirname_relative,
            push_btn=main_window.sim_conf_path_sim_exp_dir_btn,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all widget-specific undo command subclasses.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtWidgets import QUndoCommand
from betse.util.io.log import logs
from betse.util.type.types import type_check, GeneratorType
from contextlib import contextmanager

# ....................{ SUPERCLASSES                       }....................
class QBetseeWidgetUndoCommandABC(QUndoCommand):
    '''
    Abstract base class of all widget-specific undo command subclasses,
    encapsulating both the application and restoration of the contents of a
    specific type of widget.

    Attributes
    ----------
    _id : int
        Integer uniquely identifying the concrete subclass implementing this
        abstract base class of this undo command.
    _widget : QBetseeEditWidgetMixin
        Application-specific widget operated upon by this undo command.
    _synopsis : str
        Human-readable string synopsizing the operation performed by this
        undo command, preferably as a single translated sentence fragment.
        This string is identical to that returned by the :meth:`actionString`
        method, but is stored as an instance variable purely for readability.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        # Avoid circular import dependencies.
        widget: 'betsee.util.widget.abc.guiwdgabc.QBetseeEditWidgetMixin',
        synopsis: str,
    ) -> None:
        '''
        Initialize this undo command.

        Parameters
        ----------
        widget : QBetseeEditWidgetMixin
            Application-specific widget operated upon by this undo command.
        synopsis : str
            Human-readable string synopsizing the operation performed by this
            undo command, preferably as a single translated sentence fragment.
        '''

        # Initialize our superclass with the passed synopsis.
        super().__init__(synopsis)

        # Classify all parameters.
        self._synopsis = synopsis
        self._widget = widget

        # Integer uniquely identifying this concrete subclass.
        self._id = id(type(self))

    # ..................{ SUPERCLASS ~ mandatory             }..................
    # Abstract superclass methods required to be defined by each subclass.

    def undo(self) -> None:

        # Log this undo.
        logs.log_debug(
            'Undoing %s for widget "%s"...',
            self._synopsis, self._widget.obj_name)


    def redo(self) -> None:

        # Log this redo.
        logs.log_debug(
            'Redoing %s for widget "%s"...',
            self._synopsis, self._widget.obj_name)

    # ..................{ SUPERCLASS ~ optional              }..................
    # Optional superclass methods permitted to be redefined by each subclass.

    #FIXME: PySide2 should probably provide a default implementation for this
    #method identical to that defined below. Consider submitting this upstream.
    def id(self) -> int:
        '''
        Integer uniquely identifying the concrete subclass implementing this
        abstract base class.

        Our pure-C++ :class:`QUndoCommand` superclass requires this integer to
        transparently support **command compression** (i.e., automatic merging
        of adjacent undo commands of the same type)..
        '''

        return self._id

    # ..................{ CONTEXTS                           }..................
    @contextmanager
    def _in_undo_cmd(self) -> GeneratorType:
        '''
        Context manager notifying the widget associated with this undo command
        that this undo command is now being applied to it for the duration of
        this context.

        This context manager enables and then guaranteeably disables the
        :attr:`QBetseeEditWidgetMixin.is_undo_cmd_pushable` boolean even when
        fatal exceptions are raised, preventing :attr:`QBetseeEditWidgetMixin`
        subclass slots from recursively pushing additional undo commands onto
        the undo stack when already applying an undo command. Why? Because doing
        so induces infinite recursion, which is bad.

        Returns
        -----------
        contextlib._GeneratorContextManager
            Context manager notifying this widget as described above.

        Yields
        -----------
        None
            Since this context manager yields no values, the caller's ``with``
            statement must be suffixed by *no* ``as`` clause.
        '''

        # Prior state of this boolean, preserved to permit restoration.
        is_undo_cmd_pushable_prior = self._widget.is_undo_cmd_pushable

        # Yield control to the body of the caller's "with" block.
        try:
            self._widget.is_undo_cmd_pushable = False
            yield
        # Restore this boolean's state even if that block raised an exception.
        finally:
            self._widget.is_undo_cmd_pushable = is_undo_cmd_pushable_prior

# ....................{ PLACEHOLDERS                       }....................
class QBetseeUndoCommandNull(QUndoCommand):
    '''
    Placeholder undo command intended solely to simplify testing.
    '''

    # ..................{ SUPERCLASS ~ mandatory             }..................
    # Abstract superclass methods required to be defined by each subclass.

    def undo(self) -> None:

        # Log this undo.
        logs.log_debug('Undoing %s...', self.actionText())


    def redo(self) -> None:

        # Log this redo.
        logs.log_debug('Redoing %s...', self.actionText())
